# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, List, NewType, Union

JSON = NewType("JSON", Union[str, int, float, bool, None, Dict[str, "JSON"], List["JSON"]])
JSON.__doc__ = "A union type that covers all JSON-serializable data."

UserID = NewType("UserID", str)
UserID.__doc__ = "A Matrix user ID (``@user:example.com``)"
EventID = NewType("EventID", str)
EventID.__doc__ = "A Matrix event ID (``$base64`` or ``$legacyid:example.com``)"
RoomID = NewType("RoomID", str)
RoomID.__doc__ = "An internal Matrix room ID (``!randomstring:example.com``)"
RoomAlias = NewType("RoomAlias", str)
RoomAlias.__doc__ = "A Matrix room address (``#alias:example.com``)"

FilterID = NewType("FilterID", str)
FilterID.__doc__ = """
A filter ID returned by ``POST /filter`` (:meth:`mautrix.client.ClientAPI.create_filter`)
"""

BatchID = NewType("BatchID", str)
BatchID.__doc__ = """
A message batch ID returned by ``POST /batch_send`` (:meth:`mautrix.appservice.IntentAPI.batch_send`)
"""

ContentURI = NewType("ContentURI", str)
ContentURI.__doc__ = """
A Matrix `content URI`_, used by the content repository.

.. _content URI:
    https://spec.matrix.org/v1.2/client-server-api/#matrix-content-mxc-uris
"""

SyncToken = NewType("SyncToken", str)
SyncToken.__doc__ = """
A ``next_batch`` token from a ``/sync`` response (:meth:`mautrix.client.ClientAPI.sync`)
"""

DeviceID = NewType("DeviceID", str)
DeviceID.__doc__ = "A Matrix device ID. Arbitrary, potentially client-specified string."
SessionID = NewType("SessionID", str)
SessionID.__doc__ = """
A `Megolm`_ session ID.

.. _Megolm:
    https://gitlab.matrix.org/matrix-org/olm/-/blob/master/docs/megolm.md
"""
SigningKey = NewType("SigningKey", str)
SigningKey.__doc__ = "A ed25519 public key as unpadded base64"
IdentityKey = NewType("IdentityKey", str)
IdentityKey.__doc__ = "A curve25519 public key as unpadded base64"
Signature = NewType("Signature", str)
Signature.__doc__ = "An ed25519 signature as unpadded base64"



# Copyright © 2019 Damir Jelić <poljar@termina.org.uk> (under the Apache 2.0 license)
# Copyright © 2019 miruka <miruka@disroot.org> (under the Apache 2.0 license)
# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from mautrix.types import EncryptedFile

from .async_attachments import async_encrypt_attachment, async_inplace_encrypt_attachment
from .attachments import decrypt_attachment

try:
    from Crypto import Random
except ImportError:
    from Cryptodome import Random


async def _get_data_cypher_keys(data: bytes) -> tuple[bytes, EncryptedFile]:
    *chunks, keys = [i async for i in async_encrypt_attachment(data)]
    return b"".join(chunks), keys


async def test_async_encrypt():
    data = b"Test bytes"

    cyphertext, keys = await _get_data_cypher_keys(data)

    plaintext = decrypt_attachment(cyphertext, keys.key.key, keys.hashes["sha256"], keys.iv)

    assert data == plaintext


async def test_async_inplace_encrypt():
    orig_data = b"Test bytes"
    data = bytearray(orig_data)

    keys = await async_inplace_encrypt_attachment(data)

    assert data != orig_data

    decrypt_attachment(data, keys.key.key, keys.hashes["sha256"], keys.iv, inplace=True)

    assert data == orig_data



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from mautrix.client import Client
from mautrix.types import EventID

from ..custom_puppet import AutologinError, CustomPuppetError, InvalidAccessToken
from .handler import SECTION_AUTH, CommandEvent, command_handler


@command_handler(
    needs_auth=True,
    management_only=True,
    help_args="<_access token_>",
    help_section=SECTION_AUTH,
    help_text="Enable double puppeting.",
)
async def login_matrix(evt: CommandEvent) -> None:
    if len(evt.args) == 0:
        await evt.reply("**Usage:** `$cmdprefix+sp login-matrix <access token>`")
        return
    try:
        puppet = await evt.sender.get_puppet()
    except NotImplementedError:
        await evt.reply("This bridge has not implemented the login-matrix command.")
        return
    _, homeserver = Client.parse_user_id(evt.sender.mxid)
    try:
        await puppet.switch_mxid(evt.args[0], evt.sender.mxid)
        await evt.reply("Successfully enabled double puppeting.")
    except AutologinError as e:
        await evt.reply(f"Failed to create an access token: {e}")
    except CustomPuppetError as e:
        await evt.reply(str(e))


@command_handler(
    needs_auth=True,
    management_only=True,
    help_section=SECTION_AUTH,
    help_text="Disable double puppeting.",
)
async def logout_matrix(evt: CommandEvent) -> EventID:
    try:
        puppet = await evt.sender.get_puppet()
    except NotImplementedError:
        return await evt.reply("This bridge has not implemented the logout-matrix command.")
    if not puppet or not puppet.is_real_user:
        return await evt.reply("You don't have double puppeting enabled.")
    await puppet.switch_mxid(None, None)
    return await evt.reply("Successfully disabled double puppeting.")


@command_handler(
    needs_auth=True,
    help_section=SECTION_AUTH,
    help_text="Pings the Matrix server with the double puppet.",
)
async def ping_matrix(evt: CommandEvent) -> EventID:
    try:
        puppet = await evt.sender.get_puppet()
    except NotImplementedError:
        return await evt.reply("This bridge has not implemented the ping-matrix command.")
    if not puppet.is_real_user:
        return await evt.reply("You are not logged in with your Matrix account.")
    try:
        await puppet.start()
    except InvalidAccessToken:
        return await evt.reply("Your access token is invalid.")
    return await evt.reply("Your Matrix login is working.")


@command_handler(
    needs_auth=True,
    help_section=SECTION_AUTH,
    help_text="Clear the Matrix sync token stored for your double puppet.",
)
async def clear_cache_matrix(evt: CommandEvent) -> EventID:
    try:
        puppet = await evt.sender.get_puppet()
    except NotImplementedError:
        return await evt.reply("This bridge has not implemented the clear-cache-matrix command")
    if not puppet.is_real_user:
        return await evt.reply("You are not logged in with your Matrix account.")
    try:
        puppet.stop()
        puppet.next_batch = None
        await puppet.start()
    except InvalidAccessToken:
        return await evt.reply("Your access token is invalid.")
    return await evt.reply("Cleared cache successfully.")



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, List, NamedTuple, Optional, Union
from enum import IntEnum
import re

from attr import dataclass
import attr

from . import JSON
from .util import Serializable, SerializableAttrs


class VersionFormat(IntEnum):
    UNKNOWN = -1
    LEGACY = 0
    MODERN = 1

    def __repr__(self) -> str:
        return f"VersionFormat.{self.name}"


legacy_version_regex = re.compile(r"^r(\d+)\.(\d+)\.(\d+)$")
modern_version_regex = re.compile(r"^v(\d+)\.(\d+)$")


@attr.dataclass(frozen=True)
class Version(Serializable):
    format: VersionFormat
    major: int
    minor: int
    patch: int
    raw: str

    def __str__(self) -> str:
        if self.format == VersionFormat.MODERN:
            return f"v{self.major}.{self.minor}"
        elif self.format == VersionFormat.LEGACY:
            return f"r{self.major}.{self.minor}.{self.patch}"
        else:
            return self.raw

    def serialize(self) -> JSON:
        return str(self)

    @classmethod
    def deserialize(cls, raw: JSON) -> "Version":
        assert isinstance(raw, str), "versions must be strings"
        if modern := modern_version_regex.fullmatch(raw):
            major, minor = modern.groups()
            return Version(VersionFormat.MODERN, int(major), int(minor), 0, raw)
        elif legacy := legacy_version_regex.fullmatch(raw):
            major, minor, patch = legacy.groups()
            return Version(VersionFormat.LEGACY, int(major), int(minor), int(patch), raw)
        else:
            return Version(VersionFormat.UNKNOWN, 0, 0, 0, raw)


class SpecVersions:
    R010 = Version.deserialize("r0.1.0")
    R020 = Version.deserialize("r0.2.0")
    R030 = Version.deserialize("r0.3.0")
    R040 = Version.deserialize("r0.4.0")
    R050 = Version.deserialize("r0.5.0")
    R060 = Version.deserialize("r0.6.0")
    R061 = Version.deserialize("r0.6.1")
    V11 = Version.deserialize("v1.1")
    V12 = Version.deserialize("v1.2")
    V13 = Version.deserialize("v1.3")


@dataclass
class VersionsResponse(SerializableAttrs):
    versions: List[Version]
    unstable_features: Dict[str, bool] = attr.ib(factory=lambda: {})

    def supports(self, thing: Union[Version, str]) -> Optional[bool]:
        """
        Check if the versions response contains the given spec version or unstable feature.

        Args:
            thing: The spec version (as a :class:`Version` or string)
                   or unstable feature name (as a string) to check.

        Returns:
            ``True`` if the exact version or unstable feature is supported,
            ``False`` if it's not supported,
            ``None`` for unstable features which are not included in the response at all.
        """
        if isinstance(thing, Version):
            return thing in self.versions
        elif (parsed_version := Version.deserialize(thing)).format != VersionFormat.UNKNOWN:
            return parsed_version in self.versions
        return self.unstable_features.get(thing)

    def supports_at_least(self, version: Union[Version, str]) -> bool:
        """
        Check if the versions response contains the given spec version or any higher version.

        Args:
            version: The spec version as a :class:`Version` or a string.

        Returns:
            ``True`` if a version equal to or higher than the given version is found,
            ``False`` otherwise.
        """
        if isinstance(version, str):
            version = Version.deserialize(version)
        return any(v for v in self.versions if v > version)

    @property
    def latest_version(self) -> Version:
        return max(self.versions)

    @property
    def has_legacy_versions(self) -> bool:
        """
        Check if the response contains any legacy (r0.x.y) versions.

        .. deprecated:: 0.16.10
           :meth:`supports_at_least` and :meth:`supports` methods are now preferred.
        """
        return any(v for v in self.versions if v.format == VersionFormat.LEGACY)

    @property
    def has_modern_versions(self) -> bool:
        """
        Check if the response contains any modern (v1.1 or higher) versions.

        .. deprecated:: 0.16.10
           :meth:`supports_at_least` and :meth:`supports` methods are now preferred.
        """
        return self.supports_at_least(SpecVersions.V11)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Optional
import asyncio
import logging

from mautrix import client as cli
from mautrix.types import (
    ASToDeviceEvent,
    DecryptedOlmEvent,
    DeviceID,
    DeviceLists,
    DeviceOTKCount,
    EncryptionAlgorithm,
    EventType,
    Membership,
    StateEvent,
    ToDeviceEvent,
    TrustState,
    UserID,
)
from mautrix.util.logging import TraceLogger

from .account import OlmAccount
from .decrypt_megolm import MegolmDecryptionMachine
from .encrypt_megolm import MegolmEncryptionMachine
from .key_request import KeyRequestingMachine
from .key_share import KeySharingMachine
from .store import CryptoStore, StateStore
from .unwedge import OlmUnwedgingMachine


class OlmMachine(
    MegolmEncryptionMachine,
    MegolmDecryptionMachine,
    OlmUnwedgingMachine,
    KeySharingMachine,
    KeyRequestingMachine,
):
    """
    OlmMachine is the main class for handling things related to Matrix end-to-end encryption with
    Olm and Megolm. Users primarily need :meth:`encrypt_megolm_event`, :meth:`share_group_session`,
    and :meth:`decrypt_megolm_event`. Tracking device lists, establishing Olm sessions and handling
    Megolm group sessions is handled internally.
    """

    client: cli.Client
    log: TraceLogger
    crypto_store: CryptoStore
    state_store: StateStore

    account: Optional[OlmAccount]

    def __init__(
        self,
        client: cli.Client,
        crypto_store: CryptoStore,
        state_store: StateStore,
        log: Optional[TraceLogger] = None,
    ) -> None:
        super().__init__()
        self.client = client
        self.log = log or logging.getLogger("mau.crypto")
        self.crypto_store = crypto_store
        self.state_store = state_store
        self.account = None

        self.send_keys_min_trust = TrustState.UNVERIFIED
        self.share_keys_min_trust = TrustState.CROSS_SIGNED_TOFU
        self.allow_key_share = self.default_allow_key_share

        self._fetch_keys_lock = asyncio.Lock()
        self._key_request_waiters = {}
        self._inbound_session_waiters = {}
        self._prev_unwedge = {}
        self._cs_fetch_attempted = set()

        self.client.add_event_handler(
            cli.InternalEventType.DEVICE_OTK_COUNT, self.handle_otk_count, wait_sync=True
        )
        self.client.add_event_handler(cli.InternalEventType.DEVICE_LISTS, self.handle_device_lists)
        self.client.add_event_handler(EventType.TO_DEVICE_ENCRYPTED, self.handle_to_device_event)
        self.client.add_event_handler(EventType.ROOM_KEY_REQUEST, self.handle_room_key_request)
        # self.client.add_event_handler(EventType.ROOM_KEY_WITHHELD, self.handle_room_key_withheld)
        # self.client.add_event_handler(EventType.ORG_MATRIX_ROOM_KEY_WITHHELD,
        #                               self.handle_room_key_withheld)
        self.client.add_event_handler(EventType.ROOM_MEMBER, self.handle_member_event)

    async def load(self) -> None:
        """Load the Olm account into memory, or create one if the store doesn't have one stored."""
        self.account = await self.crypto_store.get_account()
        if self.account is None:
            self.account = OlmAccount()
            await self.crypto_store.put_account(self.account)

    async def handle_as_otk_counts(
        self, otk_counts: dict[UserID, dict[DeviceID, DeviceOTKCount]]
    ) -> None:
        for user_id, devices in otk_counts.items():
            for device_id, count in devices.items():
                if user_id == self.client.mxid and device_id == self.client.device_id:
                    await self.handle_otk_count(count)
                else:
                    self.log.warning(f"Got OTK count for unknown device {user_id}/{device_id}")

    async def handle_as_device_lists(self, device_lists: DeviceLists) -> None:
        asyncio.create_task(self.handle_device_lists(device_lists))

    async def handle_as_to_device_event(self, evt: ASToDeviceEvent) -> None:
        if evt.to_user_id != self.client.mxid or evt.to_device_id != self.client.device_id:
            self.log.warning(
                f"Got to-device event for unknown device {evt.to_user_id}/{evt.to_device_id}"
            )
            return
        if evt.type == EventType.TO_DEVICE_ENCRYPTED:
            await self.handle_to_device_event(evt)
        elif evt.type == EventType.ROOM_KEY_REQUEST:
            await self.handle_room_key_request(evt)
        else:
            self.log.debug(f"Got unknown to-device event {evt.type} from {evt.sender}")

    async def handle_otk_count(self, otk_count: DeviceOTKCount) -> None:
        """
        Handle the ``device_one_time_keys_count`` data in a sync response.

        This is automatically registered as an event handler and therefore called if the client you
        passed to the OlmMachine is syncing. You shouldn't need to call this yourself unless you
        do syncing in some manual way.
        """
        if otk_count.signed_curve25519 < self.account.max_one_time_keys // 2:
            self.log.debug(
                f"Sync response said we have {otk_count.signed_curve25519} signed"
                " curve25519 keys left, sharing new ones..."
            )
            await self.share_keys(otk_count.signed_curve25519)

    async def handle_device_lists(self, device_lists: DeviceLists) -> None:
        """
        Handle the ``device_lists`` data in a sync response.

        This is automatically registered as an event handler and therefore called if the client you
        passed to the OlmMachine is syncing. You shouldn't need to call this yourself unless you
        do syncing in some manual way.
        """
        if len(device_lists.changed) > 0:
            async with self._fetch_keys_lock:
                await self._fetch_keys(device_lists.changed, include_untracked=False)

    async def handle_member_event(self, evt: StateEvent) -> None:
        """
        Handle a new member event.

        This is automatically registered as an event handler and therefore called if the client you
        passed to the OlmMachine is syncing. You shouldn't need to call this yourself unless you
        receive events in some manual way (e.g. through appservice transactions)
        """
        if not await self.state_store.is_encrypted(evt.room_id):
            return
        prev = evt.prev_content.membership
        cur = evt.content.membership
        ignored_changes = {
            Membership.INVITE: Membership.JOIN,
            Membership.BAN: Membership.LEAVE,
            Membership.LEAVE: Membership.BAN,
        }
        if prev == cur or ignored_changes.get(prev) == cur:
            return
        self.log.debug(
            f"Got membership state event in {evt.room_id} changing {evt.state_key} from "
            f"{prev} to {cur}, invalidating group session"
        )
        await self.crypto_store.remove_outbound_group_session(evt.room_id)

    async def handle_to_device_event(self, evt: ToDeviceEvent) -> None:
        """
        Handle an encrypted to-device event.

        This is automatically registered as an event handler and therefore called if the client you
        passed to the OlmMachine is syncing. You shouldn't need to call this yourself unless you
        do syncing in some manual way.
        """
        self.log.trace(
            f"Handling encrypted to-device event from {evt.content.sender_key} ({evt.sender})"
        )
        decrypted_evt = await self._decrypt_olm_event(evt)
        if decrypted_evt.type == EventType.ROOM_KEY:
            await self._receive_room_key(decrypted_evt)
        elif decrypted_evt.type == EventType.FORWARDED_ROOM_KEY:
            await self._receive_forwarded_room_key(decrypted_evt)

    async def _receive_room_key(self, evt: DecryptedOlmEvent) -> None:
        # TODO nio had a comment saying "handle this better"
        #      for the case where evt.Keys.Ed25519 is none?
        if evt.content.algorithm != EncryptionAlgorithm.MEGOLM_V1 or not evt.keys.ed25519:
            return
        await self._create_group_session(
            evt.sender_key,
            evt.keys.ed25519,
            evt.content.room_id,
            evt.content.session_id,
            evt.content.session_key,
        )

    async def share_keys(self, current_otk_count: int) -> None:
        """
        Share any keys that need to be shared. This is automatically called from
        :meth:`handle_otk_count`, so you should not need to call this yourself.

        Args:
            current_otk_count: The current number of signed curve25519 keys present on the server.
        """
        device_keys = (
            self.account.get_device_keys(self.client.mxid, self.client.device_id)
            if not self.account.shared
            else None
        )
        one_time_keys = self.account.get_one_time_keys(
            self.client.mxid, self.client.device_id, current_otk_count
        )
        if not device_keys and not one_time_keys:
            self.log.warning("No one-time keys nor device keys got when trying to share keys")
            return
        if device_keys:
            self.log.debug("Going to upload initial account keys")
        self.log.debug(f"Uploading {len(one_time_keys)} one-time keys")
        await self.client.upload_keys(one_time_keys=one_time_keys, device_keys=device_keys)
        self.account.shared = True
        await self.crypto_store.put_account(self.account)
        self.log.debug("Shared keys and saved account")



from .account import OlmAccount
from .key_share import RejectKeyShare
from .sessions import InboundGroupSession, OutboundGroupSession, Session

# These have to be last
from .store import (  # isort: skip
    CryptoStore,
    MemoryCryptoStore,
    PgCryptoStateStore,
    PgCryptoStore,
    StateStore,
)

from .machine import OlmMachine  # isort: skip

__all__ = [
    "OlmAccount",
    "RejectKeyShare",
    "InboundGroupSession",
    "OutboundGroupSession",
    "Session",
    "CryptoStore",
    "MemoryCryptoStore",
    "PgCryptoStateStore",
    "PgCryptoStore",
    "StateStore",
    "OlmMachine",
    "attachments",
]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# Partly based on github.com/Cadair/python-appservice-framework (MIT license)
from __future__ import annotations

from typing import Awaitable, Callable, Optional
import asyncio
import logging

from aiohttp import web
import aiohttp

from mautrix.types import JSON, RoomAlias, UserID
from mautrix.util.logging import TraceLogger

from ..api import HTTPAPI
from .api import AppServiceAPI, IntentAPI
from .as_handler import AppServiceServerMixin
from .state_store import ASStateStore, FileASStateStore

try:
    import ssl
except ImportError:
    ssl = None

QueryFunc = Callable[[web.Request], Awaitable[Optional[web.Response]]]


class AppService(AppServiceServerMixin):
    """The main AppService container."""

    server: str
    domain: str
    id: str
    verify_ssl: bool
    tls_cert: str
    tls_key: str
    as_token: str
    hs_token: str
    bot_mxid: UserID
    default_ua: str
    default_http_retry_count: int
    bridge_name: str | None
    state_store: ASStateStore

    transactions: set[str]

    query_user: Callable[[UserID], JSON]
    query_alias: Callable[[RoomAlias], JSON]
    ready: bool
    live: bool

    loop: asyncio.AbstractEventLoop
    log: TraceLogger
    app: web.Application
    runner: web.AppRunner

    def __init__(
        self,
        server: str,
        domain: str,
        as_token: str,
        hs_token: str,
        bot_localpart: str,
        id: str,
        loop: asyncio.AbstractEventLoop | None = None,
        log: logging.Logger | str | None = None,
        verify_ssl: bool = True,
        tls_cert: str | None = None,
        tls_key: str | None = None,
        query_user: QueryFunc = None,
        query_alias: QueryFunc = None,
        bridge_name: str | None = None,
        state_store: ASStateStore = None,
        aiohttp_params: dict = None,
        ephemeral_events: bool = False,
        encryption_events: bool = False,
        default_ua: str = HTTPAPI.default_ua,
        default_http_retry_count: int = 0,
        connection_limit: int | None = None,
    ) -> None:
        super().__init__(ephemeral_events=ephemeral_events, encryption_events=encryption_events)
        self.server = server
        self.domain = domain
        self.id = id
        self.verify_ssl = verify_ssl
        self.tls_cert = tls_cert
        self.tls_key = tls_key
        self.connection_limit = connection_limit or 100
        self.as_token = as_token
        self.hs_token = hs_token
        self.bot_mxid = UserID(f"@{bot_localpart}:{domain}")
        self.default_ua = default_ua
        self.default_http_retry_count = default_http_retry_count
        self.bridge_name = bridge_name
        if not state_store:
            file = state_store if isinstance(state_store, str) else "mx-state.json"
            self.state_store = FileASStateStore(path=file, binary=False)
        elif isinstance(state_store, ASStateStore):
            self.state_store = state_store
        else:
            raise ValueError(f"Unsupported state store {type(state_store)}")

        self._http_session = None
        self._intent = None

        self.loop = loop or asyncio.get_event_loop()
        self.log = (
            logging.getLogger(log)
            if isinstance(log, str)
            else log or logging.getLogger("mau.appservice")
        )

        self.query_user = query_user or self.query_user
        self.query_alias = query_alias or self.query_alias
        self.live = True
        self.ready = False

        self.app = web.Application(loop=self.loop, **aiohttp_params if aiohttp_params else {})
        self.app.router.add_route("GET", "/_matrix/mau/live", self._liveness_probe)
        self.app.router.add_route("GET", "/_matrix/mau/ready", self._readiness_probe)
        self.register_routes(self.app)

        self.matrix_event_handler(self.state_store.update_state)

    @property
    def http_session(self) -> aiohttp.ClientSession:
        if self._http_session is None:
            raise AttributeError("the http_session attribute can only be used after starting")
        else:
            return self._http_session

    @property
    def intent(self) -> IntentAPI:
        if self._intent is None:
            raise AttributeError("the intent attribute can only be used after starting")
        else:
            return self._intent

    async def __aenter__(self) -> None:
        await self.start()

    async def __aexit__(self) -> None:
        await self.stop()

    async def start(self, host: str = "127.0.0.1", port: int = 8080) -> None:
        await self.state_store.open()
        self.log.debug(f"Starting appservice web server on {host}:{port}")
        if self.server.startswith("https://") and not self.verify_ssl:
            connector = aiohttp.TCPConnector(limit=self.connection_limit, verify_ssl=False)
        else:
            connector = aiohttp.TCPConnector(limit=self.connection_limit)
        default_headers = {"User-Agent": self.default_ua}
        self._http_session = aiohttp.ClientSession(
            loop=self.loop, connector=connector, headers=default_headers
        )
        self._intent = AppServiceAPI(
            base_url=self.server,
            bot_mxid=self.bot_mxid,
            log=self.log,
            token=self.as_token,
            state_store=self.state_store,
            bridge_name=self.bridge_name,
            client_session=self._http_session,
            default_retry_count=self.default_http_retry_count,
        ).bot_intent()
        ssl_ctx = None
        if self.tls_cert and self.tls_key:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(self.tls_cert, self.tls_key)
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        site = web.TCPSite(self.runner, host, port, ssl_context=ssl_ctx)
        await site.start()

    async def stop(self) -> None:
        self.log.debug("Stopping appservice web server")
        await self.runner.cleanup()
        self._intent = None
        await self._http_session.close()
        self._http_session = None
        await self.state_store.close()

    async def _liveness_probe(self, _: web.Request) -> web.Response:
        return web.Response(status=200 if self.live else 500, text="{}")

    async def _readiness_probe(self, _: web.Request) -> web.Response:
        return web.Response(status=200 if self.ready else 500, text="{}")



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import NamedTuple
from abc import ABC, abstractmethod

from mautrix.types import (
    CrossSigner,
    CrossSigningUsage,
    DeviceID,
    DeviceIdentity,
    EventID,
    IdentityKey,
    RoomEncryptionStateEventContent,
    RoomID,
    SessionID,
    SigningKey,
    TOFUSigningKey,
    UserID,
)

from ..account import OlmAccount
from ..sessions import InboundGroupSession, OutboundGroupSession, Session


class StateStore(ABC):
    @abstractmethod
    async def is_encrypted(self, room_id: RoomID) -> bool:
        pass

    @abstractmethod
    async def get_encryption_info(self, room_id: RoomID) -> RoomEncryptionStateEventContent | None:
        pass

    @abstractmethod
    async def find_shared_rooms(self, user_id: UserID) -> list[RoomID]:
        pass


class CryptoStore(ABC):
    """
    CryptoStore is used by :class:`OlmMachine` to store Olm and Megolm sessions, user device lists
    and message indices.
    """

    account_id: str
    """The unique identifier for the account that is stored in this CryptoStore."""
    pickle_key: str
    """The pickle key to use when pickling Olm objects."""

    @abstractmethod
    async def get_device_id(self) -> DeviceID | None:
        """
        Get the device ID corresponding to this account_id

        Returns:
            The device ID in the store.
        """

    @abstractmethod
    async def put_device_id(self, device_id: DeviceID) -> None:
        """
        Store a device ID.

        Args:
            device_id: The device ID to store.
        """

    async def open(self) -> None:
        """
        Open the store. If the store doesn't require opening any resources beforehand or only opens
        when flushing, this can be a no-op
        """

    async def close(self) -> None:
        """
        Close the store when it will no longer be used. The default implementation will simply call
        .flush(). If the store doesn't keep any persistent resources, the default implementation is
        sufficient.
        """
        await self.flush()

    async def flush(self) -> None:
        """Flush the store. If all the methods persist data immediately, this can be a no-op."""

    @abstractmethod
    async def delete(self) -> None:
        """Delete the data in the store."""

    @abstractmethod
    async def put_account(self, account: OlmAccount) -> None:
        """Insert or update the OlmAccount in the store."""

    @abstractmethod
    async def get_account(self) -> OlmAccount | None:
        """Get the OlmAccount that was previously inserted with :meth:`put_account`.
        If no account has been inserted, this must return ``None``."""

    @abstractmethod
    async def has_session(self, key: IdentityKey) -> bool:
        """
        Check whether or not the store has a session for a specific device.

        Args:
            key: The curve25519 identity key of the device to check.

        Returns:
            ``True`` if the session has at least one Olm session for the given identity key,
            ``False`` otherwise.
        """

    @abstractmethod
    async def get_sessions(self, key: IdentityKey) -> list[Session]:
        """
        Get all Olm sessions in the store for the specific device.

        Args:
            key: The curve25519 identity key of the device whose sessions to get.

        Returns:
            A list of Olm sessions for the given identity key.
            If the store contains no sessions, an empty list.
        """

    @abstractmethod
    async def get_latest_session(self, key: IdentityKey) -> Session | None:
        """
        Get the Olm session with the highest session ID (lexiographically sorting) for a specific
        device. It's usually safe to return the most recently added session if sorting by session
        ID is too difficult.

        Args:
            key: The curve25519 identity key of the device whose session to get.

        Returns:
            The most recent session for the given device.
            If the store contains no sessions, ``None``.
        """

    @abstractmethod
    async def add_session(self, key: IdentityKey, session: Session) -> None:
        """
        Insert an Olm session into the store.

        Args:
            key: The curve25519 identity key of the device with whom this session was made.
            session: The session itself.
        """

    @abstractmethod
    async def update_session(self, key: IdentityKey, session: Session) -> None:
        """
        Update a session in the store. Implementations may assume that the given session was
        previously either inserted with :meth:`add_session` or fetched with either
        :meth:`get_sessions` or :meth:`get_latest_session`.

        Args:
            key: The curve25519 identity key of the device with whom this session was made.
            session: The session itself.
        """

    @abstractmethod
    async def put_group_session(
        self,
        room_id: RoomID,
        sender_key: IdentityKey,
        session_id: SessionID,
        session: InboundGroupSession,
    ) -> None:
        """
        Insert an inbound Megolm session into the store.

        Args:
            room_id: The room ID for which this session was made.
            sender_key: The curve25519 identity key of the user who made this session.
            session_id: The unique identifier for this session.
            session: The session itself.
        """

    @abstractmethod
    async def get_group_session(
        self, room_id: RoomID, session_id: SessionID
    ) -> InboundGroupSession | None:
        """
        Get an inbound Megolm group session that was previously inserted with
        :meth:`put_group_session`.

        Args:
            room_id: The room ID for which the session was made.
            session_id: The unique identifier of the session.

        Returns:
            The :class:`InboundGroupSession` object, or ``None`` if not found.
        """

    @abstractmethod
    async def has_group_session(self, room_id: RoomID, session_id: SessionID) -> bool:
        """
        Check whether or not a specific inbound Megolm session is in the store. This is used before
        importing forwarded keys.

        Args:
            room_id: The room ID for which the session was made.
            session_id: The unique identifier of the session.

        Returns:
            ``True`` if the store has a session with the given ID, ``False`` otherwise.
        """

    @abstractmethod
    async def add_outbound_group_session(self, session: OutboundGroupSession) -> None:
        """
        Insert an outbound Megolm session into the store.

        The store should index inserted sessions by the room_id field of the session to support
        getting and removing sessions. There will only be one outbound session per room ID at a
        time.

        Args:
            session: The session itself.
        """

    @abstractmethod
    async def update_outbound_group_session(self, session: OutboundGroupSession) -> None:
        """
        Update an outbound Megolm session in the store. Implementations may assume that the given
        session was previously either inserted with :meth:`add_outbound_group_session` or fetched
        with :meth:`get_outbound_group_session`.

        Args:
            session: The session itself.
        """

    @abstractmethod
    async def get_outbound_group_session(self, room_id: RoomID) -> OutboundGroupSession | None:
        """
        Get the stored outbound Megolm session from the store.

        Args:
            room_id: The room whose session to get.

        Returns:
            The :class:`OutboundGroupSession` object, or ``None`` if not found.
        """

    @abstractmethod
    async def remove_outbound_group_session(self, room_id: RoomID) -> None:
        """
        Remove the stored outbound Megolm session for a specific room.

        This is used when a membership change is received in a specific room.

        Args:
            room_id: The room whose session to remove.
        """

    @abstractmethod
    async def remove_outbound_group_sessions(self, rooms: list[RoomID]) -> None:
        """
        Remove the stored outbound Megolm session for multiple rooms.

        This is used when the device list of a user changes.

        Args:
            rooms: The list of rooms whose sessions to remove.
        """

    @abstractmethod
    async def validate_message_index(
        self,
        sender_key: IdentityKey,
        session_id: SessionID,
        event_id: EventID,
        index: int,
        timestamp: int,
    ) -> bool:
        """
        Validate that a specific message isn't a replay attack.

        Implementations should store a map from ``(sender_key, session_id, index)`` to
        ``(event_id, timestamp)``, then use that map to check whether or not the message
        index is valid:

        * If the map key doesn't exist, the given values should be stored and the message is valid.
        * If the map key exists and the stored values match the given values, the message is valid.
        * If the map key exists, but the stored values do not match the given values, the message
          is not valid.

        Args:
            sender_key: The curve25519 identity key of the user who sent the message.
            session_id: The Megolm session ID for the session with which the message was encrypted.
            event_id: The event ID of the message.
            index: The Megolm message index of the message.
            timestamp: The timestamp of the message.

        Returns:
            ``True`` if the message is valid, ``False`` if not.
        """

    @abstractmethod
    async def get_devices(self, user_id: UserID) -> dict[DeviceID, DeviceIdentity] | None:
        """
        Get all devices for a given user.

        Args:
            user_id: The ID of the user whose devices to get.

        Returns:
            If there has been a previous call to :meth:`put_devices` with the same user ID (even
            with an empty dict), a dict from device ID to :class:`DeviceIdentity` object.
            Otherwise, ``None``.
        """

    @abstractmethod
    async def get_device(self, user_id: UserID, device_id: DeviceID) -> DeviceIdentity | None:
        """
        Get a specific device identity.

        Args:
            user_id: The ID of the user whose device to get.
            device_id: The ID of the device to get.

        Returns:
            The :class:`DeviceIdentity` object, or ``None`` if not found.
        """

    @abstractmethod
    async def find_device_by_key(
        self, user_id: UserID, identity_key: IdentityKey
    ) -> DeviceIdentity | None:
        """
        Find a specific device identity based on the identity key.

        Args:
            user_id: The ID of the user whose device to get.
            identity_key: The identity key of the device to get.

        Returns:
            The :class:`DeviceIdentity` object, or ``None`` if not found.
        """

    @abstractmethod
    async def put_devices(self, user_id: UserID, devices: dict[DeviceID, DeviceIdentity]) -> None:
        """
        Replace the stored device list for a specific user.

        Args:
            user_id: The ID of the user whose device list to update.
            devices: A dict from device ID to :class:`DeviceIdentity` object. The dict may be empty.
        """

    @abstractmethod
    async def filter_tracked_users(self, users: list[UserID]) -> list[UserID]:
        """
        Filter a list of user IDs to only include users whose device lists are being tracked.

        Args:
            users: The list of user IDs to filter.

        Returns:
            A filtered version of the input list that only includes users who have had a previous
            call to :meth:`put_devices` (even if the call was with an empty dict).
        """

    @abstractmethod
    async def put_cross_signing_key(
        self, user_id: UserID, usage: CrossSigningUsage, key: SigningKey
    ) -> None:
        """
        Store a single cross-signing key.

        Args:
            user_id: The user whose cross-signing key is being stored.
            usage: The type of key being stored.
            key: The key itself.
        """

    @abstractmethod
    async def get_cross_signing_keys(
        self, user_id: UserID
    ) -> dict[CrossSigningUsage, TOFUSigningKey]:
        """
        Retrieve stored cross-signing keys for a specific user.

        Args:
            user_id: The user whose cross-signing keys to get.

        Returns:
            A map from the type of key to a tuple containing the current key and the key that was
            seen first. If the keys are different, it should be treated as a local TOFU violation.
        """

    @abstractmethod
    async def put_signature(
        self, target: CrossSigner, signer: CrossSigner, signature: str
    ) -> None:
        """
        Store a signature for a given key from a given key.

        Args:
            target: The user ID and key being signed.
            signer: The user ID and key who are doing the signing.
            signature: The signature.
        """

    @abstractmethod
    async def is_key_signed_by(self, target: CrossSigner, signer: CrossSigner) -> bool:
        """
        Check if a given key is signed by the given signer.

        Args:
            target: The key to check.
            signer: The signer who is expected to have signed the key.

        Returns:
            ``True`` if the database contains a signature for the key, ``False`` otherwise.
        """

    @abstractmethod
    async def drop_signatures_by_key(self, signer: CrossSigner) -> int:
        """
        Delete signatures made by the given key.

        Args:
            signer: The key whose signatures to delete.

        Returns:
            The number of signatures deleted.
        """



# Copyright (c) 2022 Sumner Evans
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional
import logging

from aiohttp.client import ClientTimeout
from attr import dataclass
import aiohttp

from mautrix.api import HTTPAPI
from mautrix.types import EventType, MessageType, SerializableAttrs, SerializableEnum


class MessageSendCheckpointStep(SerializableEnum):
    CLIENT = "CLIENT"
    HOMESERVER = "HOMESERVER"
    BRIDGE = "BRIDGE"
    DECRYPTED = "DECRYPTED"
    REMOTE = "REMOTE"
    COMMAND = "COMMAND"


class MessageSendCheckpointStatus(SerializableEnum):
    SUCCESS = "SUCCESS"
    WILL_RETRY = "WILL_RETRY"
    PERM_FAILURE = "PERM_FAILURE"
    UNSUPPORTED = "UNSUPPORTED"
    TIMEOUT = "TIMEOUT"


class MessageSendCheckpointReportedBy(SerializableEnum):
    ASMUX = "ASMUX"
    BRIDGE = "BRIDGE"


@dataclass
class MessageSendCheckpoint(SerializableAttrs):
    event_id: str
    room_id: str
    step: MessageSendCheckpointStep
    timestamp: int
    status: MessageSendCheckpointStatus
    event_type: EventType
    reported_by: MessageSendCheckpointReportedBy
    retry_num: int = 0
    message_type: Optional[MessageType] = None
    info: Optional[str] = None
    client_type: Optional[str] = None
    client_version: Optional[str] = None

    async def send(self, endpoint: str, as_token: str, log: logging.Logger) -> None:
        if not endpoint:
            return
        try:
            headers = {"Authorization": f"Bearer {as_token}", "User-Agent": HTTPAPI.default_ua}
            async with aiohttp.ClientSession() as sess, sess.post(
                endpoint,
                json={"checkpoints": [self.serialize()]},
                headers=headers,
                timeout=ClientTimeout(30),
            ) as resp:
                if not 200 <= resp.status < 300:
                    text = await resp.text()
                    text = text.replace("\n", "\\n")
                    log.warning(
                        f"Unexpected status code {resp.status} sending checkpoint "
                        f"for {self.event_id} ({self.step}/{self.status}): {text}"
                    )
                else:
                    log.info(
                        f"Successfully sent checkpoint for {self.event_id} "
                        f"({self.step}/{self.status})"
                    )
        except Exception as e:
            log.warning(
                f"Failed to send checkpoint for {self.event_id} ({self.step}/{self.status}): "
                f"{type(e).__name__}: {e}"
            )


CHECKPOINT_TYPES = {
    EventType.ROOM_REDACTION,
    EventType.ROOM_MESSAGE,
    EventType.ROOM_ENCRYPTED,
    EventType.ROOM_MEMBER,
    EventType.ROOM_NAME,
    EventType.ROOM_AVATAR,
    EventType.ROOM_TOPIC,
    EventType.STICKER,
    EventType.REACTION,
    EventType.CALL_INVITE,
    EventType.CALL_CANDIDATES,
    EventType.CALL_SELECT_ANSWER,
    EventType.CALL_ANSWER,
    EventType.CALL_HANGUP,
    EventType.CALL_REJECT,
    EventType.CALL_NEGOTIATE,
}



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from attr import dataclass

from .enum import ExtensibleEnum
from .serializable_attrs import SerializableAttrs


def test_extensible_enum_int():
    class Hello(ExtensibleEnum):
        HI = 1
        HMM = 2

    assert Hello.HI.value == 1
    assert Hello.HI.key == "HI"
    assert 1 in Hello
    assert Hello(1) == Hello.HI
    assert Hello["HMM"] == Hello.HMM

    assert len(Hello) == 2
    hello3 = Hello(3)
    assert hello3.value == 3
    assert not hello3.key
    Hello.YAY = hello3
    assert len(Hello) == 3
    assert hello3.key == "YAY"

    @dataclass
    class Wrapper(SerializableAttrs):
        hello: Hello

    assert Wrapper.deserialize({"hello": 1}).hello == Hello.HI
    assert Wrapper.deserialize({"hello": 2}).hello == Hello.HMM
    assert Wrapper.deserialize({"hello": 3}).hello == hello3
    assert Wrapper.deserialize({"hello": 4}).hello.value == 4


def test_extensible_enum_str():
    class Hello(ExtensibleEnum):
        HI = "hi"
        HMM = "🤔"

    assert Hello.HI.value == "hi"
    assert Hello.HI.key == "HI"
    assert "🤔" in Hello
    assert Hello("🤔") == Hello.HMM
    assert Hello["HI"] == Hello.HI

    assert len(Hello) == 2
    hello3 = Hello("yay")
    assert hello3.value == "yay"
    assert not hello3.key
    Hello.YAY = hello3
    assert len(Hello) == 3
    assert hello3.key == "YAY"

    @dataclass
    class Wrapper(SerializableAttrs):
        hello: Hello

    assert Wrapper.deserialize({"hello": "hi"}).hello == Hello.HI
    assert Wrapper.deserialize({"hello": "🤔"}).hello == Hello.HMM
    assert Wrapper.deserialize({"hello": "yay"}).hello == hello3
    assert Wrapper.deserialize({"hello": "thonk"}).hello.value == "thonk"



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .base import MatrixResponseError


class WellKnownError(MatrixResponseError):
    """
    An error that occurred during server discovery.

    https://matrix.org/docs/spec/client_server/latest#get-well-known-matrix-client
    """

    pass


class WellKnownUnexpectedStatus(WellKnownError):
    def __init__(self, status: int) -> None:
        super().__init__(f"Unexpected status code {status} when fetching .well-known file")
        self.status = status


class WellKnownNotJSON(WellKnownError):
    def __init__(self) -> None:
        super().__init__(".well-known response was not JSON")


class WellKnownMissingHomeserver(WellKnownError):
    def __init__(self) -> None:
        super().__init__("No homeserver found in .well-known response")


class WellKnownNotURL(WellKnownError):
    def __init__(self) -> None:
        super().__init__("Homeserver base URL in .well-known response was not a valid URL")


class WellKnownUnsupportedScheme(WellKnownError):
    def __init__(self, scheme: str) -> None:
        super().__init__(f"URL in .well-known response has unsupported scheme {scheme}")


class WellKnownInvalidVersionsResponse(WellKnownError):
    def __init__(self) -> None:
        super().__init__(
            "URL in .well-known response didn't respond to versions endpoint properly"
        )



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Dict
import asyncio

from mautrix.types import (
    DecryptedOlmEvent,
    DeviceID,
    DeviceIdentity,
    EncryptedOlmEventContent,
    EncryptionKeyAlgorithm,
    EventType,
    OlmEventKeys,
    ToDeviceEventContent,
    UserID,
)

from .base import BaseOlmMachine, verify_signature_json
from .sessions import Session

ClaimKeysList = Dict[UserID, Dict[DeviceID, DeviceIdentity]]


class OlmEncryptionMachine(BaseOlmMachine):
    _claim_keys_lock: asyncio.Lock
    _olm_lock: asyncio.Lock

    def __init__(self):
        self._claim_keys_lock = asyncio.Lock()
        self._olm_lock = asyncio.Lock()

    async def _encrypt_olm_event(
        self, session: Session, recipient: DeviceIdentity, event_type: EventType, content: Any
    ) -> EncryptedOlmEventContent:
        evt = DecryptedOlmEvent(
            sender=self.client.mxid,
            sender_device=self.client.device_id,
            keys=OlmEventKeys(ed25519=self.account.signing_key),
            recipient=recipient.user_id,
            recipient_keys=OlmEventKeys(ed25519=recipient.signing_key),
            type=event_type,
            content=content,
        )
        ciphertext = session.encrypt(evt.json())
        await self.crypto_store.update_session(recipient.identity_key, session)
        return EncryptedOlmEventContent(
            ciphertext={recipient.identity_key: ciphertext}, sender_key=self.account.identity_key
        )

    async def _create_outbound_sessions(
        self, users: ClaimKeysList, _force_recreate_session: bool = False
    ) -> None:
        async with self._claim_keys_lock:
            return await self._create_outbound_sessions_locked(users, _force_recreate_session)

    async def _create_outbound_sessions_locked(
        self, users: ClaimKeysList, _force_recreate_session: bool = False
    ) -> None:
        request: Dict[UserID, Dict[DeviceID, EncryptionKeyAlgorithm]] = {}
        expected_devices = set()
        for user_id, devices in users.items():
            request[user_id] = {}
            for device_id, identity in devices.items():
                if _force_recreate_session or not await self.crypto_store.has_session(
                    identity.identity_key
                ):
                    request[user_id][device_id] = EncryptionKeyAlgorithm.SIGNED_CURVE25519
                    expected_devices.add((user_id, device_id))
            if not request[user_id]:
                del request[user_id]
        if not request:
            return
        request_device_count = len(expected_devices)
        keys = await self.client.claim_keys(request)
        for server, info in (keys.failures or {}).items():
            self.log.warning(f"Key claim failure for {server}: {info}")
        for user_id, devices in keys.one_time_keys.items():
            for device_id, one_time_keys in devices.items():
                expected_devices.discard((user_id, device_id))
                key_id, one_time_key_data = one_time_keys.popitem()
                one_time_key = one_time_key_data["key"]
                identity = users[user_id][device_id]
                if not verify_signature_json(
                    one_time_key_data, user_id, device_id, identity.signing_key
                ):
                    self.log.warning(f"Invalid signature for {device_id} of {user_id}")
                else:
                    session = self.account.new_outbound_session(
                        identity.identity_key, one_time_key
                    )
                    await self.crypto_store.add_session(identity.identity_key, session)
                    self.log.debug(
                        f"Created new Olm session with {user_id}/{device_id} "
                        f"(OTK ID: {key_id})"
                    )
        if expected_devices:
            if request_device_count == 1:
                raise Exception(
                    "Key claim response didn't contain key "
                    f"for queried device {expected_devices.pop()}"
                )
            else:
                self.log.warning(
                    "Key claim response didn't contain keys for %d out of %d expected devices: %s",
                    len(expected_devices),
                    request_device_count,
                    expected_devices,
                )

    async def send_encrypted_to_device(
        self,
        device: DeviceIdentity,
        event_type: EventType,
        content: ToDeviceEventContent,
        _force_recreate_session: bool = False,
    ) -> None:
        await self._create_outbound_sessions(
            {device.user_id: {device.device_id: device}},
            _force_recreate_session=_force_recreate_session,
        )
        session = await self.crypto_store.get_latest_session(device.identity_key)
        async with self._olm_lock:
            encrypted_content = await self._encrypt_olm_event(session, device, event_type, content)
            await self.client.send_to_one_device(
                EventType.TO_DEVICE_ENCRYPTED, device.user_id, device.device_id, encrypted_content
            )



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, ClassVar, Optional

from mautrix.types import JSON, ExtensibleEnum, Serializable, SerializableEnum

class RoomType(ExtensibleEnum):
    SPACE: "RoomType"

class EventType(Serializable):
    class Class(SerializableEnum):
        UNKNOWN = "unknown"
        STATE = "state"
        MESSAGE = "message"
        ACCOUNT_DATA = "account_data"
        EPHEMERAL = "ephemeral"
        TO_DEVICE = "to_device"
    _by_event_type: ClassVar[dict[str, EventType]]

    ROOM_CANONICAL_ALIAS: "EventType"
    ROOM_CREATE: "EventType"
    ROOM_JOIN_RULES: "EventType"
    ROOM_MEMBER: "EventType"
    ROOM_POWER_LEVELS: "EventType"
    ROOM_HISTORY_VISIBILITY: "EventType"
    ROOM_NAME: "EventType"
    ROOM_TOPIC: "EventType"
    ROOM_AVATAR: "EventType"
    ROOM_PINNED_EVENTS: "EventType"
    ROOM_TOMBSTONE: "EventType"
    ROOM_ENCRYPTION: "EventType"

    SPACE_CHILD: "EventType"
    SPACE_PARENT: "EventType"

    ROOM_REDACTION: "EventType"
    ROOM_MESSAGE: "EventType"
    ROOM_ENCRYPTED: "EventType"
    STICKER: "EventType"
    REACTION: "EventType"

    CALL_INVITE: "EventType"
    CALL_CANDIDATES: "EventType"
    CALL_SELECT_ANSWER: "EventType"
    CALL_ANSWER: "EventType"
    CALL_HANGUP: "EventType"
    CALL_REJECT: "EventType"
    CALL_NEGOTIATE: "EventType"

    BEEPER_MESSAGE_STATUS: "EventType"

    RECEIPT: "EventType"
    TYPING: "EventType"
    PRESENCE: "EventType"

    DIRECT: "EventType"
    PUSH_RULES: "EventType"
    TAG: "EventType"
    IGNORED_USER_LIST: "EventType"

    TO_DEVICE_ENCRYPTED: "EventType"
    TO_DEVICE_DUMMY: "EventType"
    ROOM_KEY: "EventType"
    ROOM_KEY_WITHHELD: "EventType"
    ORG_MATRIX_ROOM_KEY_WITHHELD: "EventType"
    ROOM_KEY_REQUEST: "EventType"
    FORWARDED_ROOM_KEY: "EventType"

    ALL: "EventType"

    is_message: bool
    is_state: bool
    is_ephemeral: bool
    is_account_data: bool
    is_to_device: bool

    t: str
    t_class: Class
    def __init__(self, t: str, t_class: Class) -> None: ...
    @classmethod
    def find(cls, t: str, t_class: Optional[Class] = None) -> "EventType": ...
    def serialize(self) -> JSON: ...
    @classmethod
    def deserialize(cls, raw: JSON) -> Any: ...
    def with_class(self, t_class: Class) -> "EventType": ...



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from mautrix.client.state_store import SyncStore
from mautrix.types import (
    CrossSigner,
    CrossSigningUsage,
    DeviceID,
    DeviceIdentity,
    EventID,
    IdentityKey,
    RoomID,
    SessionID,
    SigningKey,
    SyncToken,
    TOFUSigningKey,
    UserID,
)

from ..account import OlmAccount
from ..sessions import InboundGroupSession, OutboundGroupSession, Session
from .abstract import CryptoStore


class MemoryCryptoStore(CryptoStore, SyncStore):
    _device_id: DeviceID | None
    _sync_token: SyncToken | None
    _account: OlmAccount | None
    _message_indices: dict[tuple[IdentityKey, SessionID, int], tuple[EventID, int]]
    _devices: dict[UserID, dict[DeviceID, DeviceIdentity]]
    _olm_sessions: dict[IdentityKey, list[Session]]
    _inbound_sessions: dict[tuple[RoomID, SessionID], InboundGroupSession]
    _outbound_sessions: dict[RoomID, OutboundGroupSession]
    _signatures: dict[CrossSigner, dict[CrossSigner, str]]
    _cross_signing_keys: dict[UserID, dict[CrossSigningUsage, TOFUSigningKey]]

    def __init__(self, account_id: str, pickle_key: str) -> None:
        self.account_id = account_id
        self.pickle_key = pickle_key

        self._sync_token = None
        self._device_id = None
        self._account = None
        self._message_indices = {}
        self._devices = {}
        self._olm_sessions = {}
        self._inbound_sessions = {}
        self._outbound_sessions = {}
        self._signatures = {}
        self._cross_signing_keys = {}

    async def get_device_id(self) -> DeviceID | None:
        return self._device_id

    async def put_device_id(self, device_id: DeviceID) -> None:
        self._device_id = device_id

    async def put_next_batch(self, next_batch: SyncToken) -> None:
        self._sync_token = next_batch

    async def get_next_batch(self) -> SyncToken:
        return self._sync_token

    async def delete(self) -> None:
        self._account = None
        self._device_id = None
        self._olm_sessions = {}
        self._outbound_sessions = {}

    async def put_account(self, account: OlmAccount) -> None:
        self._account = account

    async def get_account(self) -> OlmAccount:
        return self._account

    async def has_session(self, key: IdentityKey) -> bool:
        return key in self._olm_sessions

    async def get_sessions(self, key: IdentityKey) -> list[Session]:
        return self._olm_sessions.get(key, [])

    async def get_latest_session(self, key: IdentityKey) -> Session | None:
        try:
            return self._olm_sessions[key][-1]
        except (KeyError, IndexError):
            return None

    async def add_session(self, key: IdentityKey, session: Session) -> None:
        self._olm_sessions.setdefault(key, []).append(session)

    async def update_session(self, key: IdentityKey, session: Session) -> None:
        # This is a no-op as the session object is the same one previously added.
        pass

    async def put_group_session(
        self,
        room_id: RoomID,
        sender_key: IdentityKey,
        session_id: SessionID,
        session: InboundGroupSession,
    ) -> None:
        self._inbound_sessions[(room_id, session_id)] = session

    async def get_group_session(
        self, room_id: RoomID, session_id: SessionID
    ) -> InboundGroupSession:
        return self._inbound_sessions.get((room_id, session_id))

    async def has_group_session(self, room_id: RoomID, session_id: SessionID) -> bool:
        return (room_id, session_id) in self._inbound_sessions

    async def add_outbound_group_session(self, session: OutboundGroupSession) -> None:
        self._outbound_sessions[session.room_id] = session

    async def update_outbound_group_session(self, session: OutboundGroupSession) -> None:
        # This is a no-op as the session object is the same one previously added.
        pass

    async def get_outbound_group_session(self, room_id: RoomID) -> OutboundGroupSession | None:
        return self._outbound_sessions.get(room_id)

    async def remove_outbound_group_session(self, room_id: RoomID) -> None:
        self._outbound_sessions.pop(room_id, None)

    async def remove_outbound_group_sessions(self, rooms: list[RoomID]) -> None:
        for room_id in rooms:
            self._outbound_sessions.pop(room_id, None)

    async def validate_message_index(
        self,
        sender_key: IdentityKey,
        session_id: SessionID,
        event_id: EventID,
        index: int,
        timestamp: int,
    ) -> bool:
        try:
            return self._message_indices[(sender_key, session_id, index)] == (event_id, timestamp)
        except KeyError:
            self._message_indices[(sender_key, session_id, index)] = (event_id, timestamp)
            return True

    async def get_devices(self, user_id: UserID) -> dict[DeviceID, DeviceIdentity] | None:
        return self._devices.get(user_id)

    async def get_device(self, user_id: UserID, device_id: DeviceID) -> DeviceIdentity | None:
        return self._devices.get(user_id, {}).get(device_id)

    async def find_device_by_key(
        self, user_id: UserID, identity_key: IdentityKey
    ) -> DeviceIdentity | None:
        for device in self._devices.get(user_id, {}).values():
            if device.identity_key == identity_key:
                return device
        return None

    async def put_devices(self, user_id: UserID, devices: dict[DeviceID, DeviceIdentity]) -> None:
        self._devices[user_id] = devices

    async def filter_tracked_users(self, users: list[UserID]) -> list[UserID]:
        return [user_id for user_id in users if user_id in self._devices]

    async def put_cross_signing_key(
        self, user_id: UserID, usage: CrossSigningUsage, key: SigningKey
    ) -> None:
        try:
            current = self._cross_signing_keys[user_id][usage]
        except KeyError:
            self._cross_signing_keys.setdefault(user_id, {})[usage] = TOFUSigningKey(
                key=key, first=key
            )
        else:
            current.key = key

    async def get_cross_signing_keys(
        self, user_id: UserID
    ) -> dict[CrossSigningUsage, TOFUSigningKey]:
        return self._cross_signing_keys.get(user_id, {})

    async def put_signature(
        self, target: CrossSigner, signer: CrossSigner, signature: str
    ) -> None:
        self._signatures.setdefault(signer, {})[target] = signature

    async def is_key_signed_by(self, target: CrossSigner, signer: CrossSigner) -> bool:
        return target in self._signatures.get(signer, {})

    async def drop_signatures_by_key(self, signer: CrossSigner) -> int:
        deleted = self._signatures.pop(signer, None)
        return len(deleted)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import AsyncContextManager, AsyncIterator, Callable
from contextlib import asynccontextmanager
import json
import os
import pathlib
import random
import string
import time

import asyncpg
import pytest
import sqlalchemy as sql

from mautrix.types import EncryptionAlgorithm, Member, Membership, RoomID, StateEvent, UserID
from mautrix.util.async_db import Database
from mautrix.util.db import Base

from .. import MemoryStateStore, StateStore
from ..asyncpg import PgStateStore
from ..sqlalchemy import RoomState, SQLStateStore, UserProfile


@asynccontextmanager
async def async_postgres_store() -> AsyncIterator[PgStateStore]:
    try:
        pg_url = os.environ["MEOW_TEST_PG_URL"]
    except KeyError:
        pytest.skip("Skipped Postgres tests (MEOW_TEST_PG_URL not specified)")
        return
    conn: asyncpg.Connection = await asyncpg.connect(pg_url)
    schema_name = "".join(random.choices(string.ascii_lowercase, k=8))
    schema_name = f"test_schema_{schema_name}_{int(time.time())}"
    await conn.execute(f"CREATE SCHEMA {schema_name}")
    db = Database.create(
        pg_url,
        upgrade_table=PgStateStore.upgrade_table,
        db_args={"min_size": 1, "max_size": 3, "server_settings": {"search_path": schema_name}},
    )
    store = PgStateStore(db)
    await db.start()
    yield store
    await db.stop()
    await conn.execute(f"DROP SCHEMA {schema_name} CASCADE")
    await conn.close()


@asynccontextmanager
async def async_sqlite_store() -> AsyncIterator[PgStateStore]:
    db = Database.create(
        "sqlite:///:memory:", upgrade_table=PgStateStore.upgrade_table, db_args={"min_size": 1}
    )
    store = PgStateStore(db)
    await db.start()
    yield store
    await db.stop()


@asynccontextmanager
async def alchemy_store() -> AsyncIterator[SQLStateStore]:
    db = sql.create_engine("sqlite:///:memory:")
    Base.metadata.bind = db
    for table in (RoomState, UserProfile):
        table.bind(db)
    Base.metadata.create_all()
    yield SQLStateStore()
    db.dispose()


@asynccontextmanager
async def memory_store() -> AsyncIterator[MemoryStateStore]:
    yield MemoryStateStore()


@pytest.fixture(params=[async_postgres_store, async_sqlite_store, alchemy_store, memory_store])
async def store(request) -> AsyncIterator[StateStore]:
    param: Callable[[], AsyncContextManager[StateStore]] = request.param
    async with param() as state_store:
        yield state_store


def read_state_file(request, file) -> dict[RoomID, list[StateEvent]]:
    path = pathlib.Path(request.node.fspath).with_name(file)
    with path.open() as fp:
        content = json.load(fp)
    return {
        room_id: [StateEvent.deserialize({**evt, "room_id": room_id}) for evt in events]
        for room_id, events in content.items()
    }


async def store_room_state(request, store: StateStore) -> None:
    room_state_changes = read_state_file(request, "new_state.json")
    for events in room_state_changes.values():
        for evt in events:
            await store.update_state(evt)


async def get_all_members(request, store: StateStore) -> None:
    room_state = read_state_file(request, "members.json")
    for room_id, member_events in room_state.items():
        await store.set_members(room_id, {evt.state_key: evt.content for evt in member_events})


async def get_joined_members(request, store: StateStore) -> None:
    path = pathlib.Path(request.node.fspath).with_name("joined_members.json")
    with path.open() as fp:
        content = json.load(fp)
    for room_id, members in content.items():
        parsed_members = {
            user_id: Member(
                membership=Membership.JOIN,
                displayname=member.get("display_name", ""),
                avatar_url=member.get("avatar_url", ""),
            )
            for user_id, member in members.items()
        }
        await store.set_members(room_id, parsed_members, only_membership=Membership.JOIN)


async def test_basic(store: StateStore) -> None:
    room_id = RoomID("!foo:example.com")
    user_id = UserID("@tulir:example.com")

    assert not await store.is_encrypted(room_id)
    assert not await store.is_joined(room_id, user_id)
    await store.joined(room_id, user_id)
    assert await store.is_joined(room_id, user_id)

    assert not await store.has_encryption_info_cached(RoomID("!unknown-room:example.com"))
    assert await store.is_encrypted(RoomID("!unknown-room:example.com")) is None


async def test_basic_updated(request, store: StateStore) -> None:
    await store_room_state(request, store)
    test_group = RoomID("!telegram-group:example.com")
    assert await store.is_encrypted(test_group)
    assert (await store.get_encryption_info(test_group)).algorithm == EncryptionAlgorithm.MEGOLM_V1
    assert not await store.is_encrypted(RoomID("!unencrypted-room:example.com"))


async def test_updates(request, store: StateStore) -> None:
    await store_room_state(request, store)
    room_id = RoomID("!telegram-group:example.com")
    initial_members = {"@tulir:example.com", "@telegram_84359547:example.com"}
    joined_members = initial_members | {
        "@telegrambot:example.com",
        "@telegram_5647382910:example.com",
        "@telegram_374880943:example.com",
        "@telegram_987654321:example.com",
        "@telegram_123456789:example.com",
    }
    left_members = {"@telegram_476034259:example.com", "@whatsappbot:example.com"}
    full_members = joined_members | left_members
    any_membership = (
        Membership.JOIN,
        Membership.INVITE,
        Membership.LEAVE,
        Membership.BAN,
        Membership.KNOCK,
    )
    leave_memberships = (Membership.BAN, Membership.LEAVE)
    assert set(await store.get_members(room_id)) == initial_members
    await get_all_members(request, store)
    assert set(await store.get_members(room_id)) == joined_members
    assert set(await store.get_members(room_id, memberships=any_membership)) == full_members
    await get_joined_members(request, store)
    assert set(await store.get_members(room_id)) == joined_members
    assert set(await store.get_members(room_id, memberships=any_membership)) == full_members
    assert set(await store.get_members(room_id, memberships=leave_memberships)) == left_members
    assert set(
        await store.get_members_filtered(
            room_id,
            memberships=leave_memberships,
            not_id="",
            not_prefix="@telegram_",
            not_suffix=":example.com",
        )
    ) == {"@whatsappbot:example.com"}



from ..util.async_getter_lock import async_getter_lock
from .bridge import Bridge, HomeserverSoftware
from .config import BaseBridgeConfig
from .custom_puppet import (
    AutologinError,
    CustomPuppetError,
    CustomPuppetMixin,
    EncryptionKeysFound,
    HomeserverURLNotFound,
    InvalidAccessToken,
    OnlyLoginSelf,
    OnlyLoginTrustedDomain,
)
from .disappearing_message import AbstractDisappearingMessage
from .matrix import BaseMatrixHandler
from .notification_disabler import NotificationDisabler
from .portal import BasePortal, DMCreateError, IgnoreMatrixInvite, RejectMatrixInvite
from .puppet import BasePuppet
from .user import BaseUser

__all__ = [
    "async_getter_lock",
    "Bridge",
    "HomeserverSoftware",
    "BaseBridgeConfig",
    "AutologinError",
    "CustomPuppetError",
    "CustomPuppetMixin",
    "HomeserverURLNotFound",
    "InvalidAccessToken",
    "OnlyLoginSelf",
    "OnlyLoginTrustedDomain",
    "AbstractDisappearingMessage",
    "BaseMatrixHandler",
    "NotificationDisabler",
    "BasePortal",
    "BasePuppet",
    "BaseUser",
    "state_store",
    "commands",
]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Awaitable, Callable
from abc import ABC

from mautrix import __optional_imports__
from mautrix.bridge.portal import BasePortal
from mautrix.crypto import StateStore
from mautrix.types import RoomEncryptionStateEventContent, RoomID, UserID
from mautrix.util.async_db import Database

GetPortalFunc = Callable[[RoomID], Awaitable[BasePortal]]


class BaseCryptoStateStore(StateStore, ABC):
    get_portal: GetPortalFunc

    def __init__(self, get_portal: GetPortalFunc):
        self.get_portal = get_portal

    async def is_encrypted(self, room_id: RoomID) -> bool:
        portal = await self.get_portal(room_id)
        return portal.encrypted if portal else False


try:
    from mautrix.client.state_store.sqlalchemy import RoomState, UserProfile

    class SQLCryptoStateStore(BaseCryptoStateStore):
        @staticmethod
        async def find_shared_rooms(user_id: UserID) -> list[RoomID]:
            return [profile.room_id for profile in UserProfile.find_rooms_with_user(user_id)]

        @staticmethod
        async def get_encryption_info(room_id: RoomID) -> RoomEncryptionStateEventContent | None:
            state = RoomState.get(room_id)
            if not state:
                return None
            return state.encryption

except ImportError:
    if __optional_imports__:
        raise
    UserProfile = None
    RoomState = None
    SQLCryptoStateStore = None


class PgCryptoStateStore(BaseCryptoStateStore):
    db: Database

    def __init__(self, db: Database, get_portal: GetPortalFunc) -> None:
        super().__init__(get_portal)
        self.db = db

    async def find_shared_rooms(self, user_id: UserID) -> list[RoomID]:
        rows = await self.db.fetch(
            "SELECT room_id FROM mx_user_profile "
            "LEFT JOIN portal ON portal.mxid=mx_user_profile.room_id "
            "WHERE user_id=$1 AND portal.encrypted=true",
            user_id,
        )
        return [row["room_id"] for row in rows]

    async def get_encryption_info(self, room_id: RoomID) -> RoomEncryptionStateEventContent | None:
        val = await self.db.fetchval(
            "SELECT encryption FROM mx_room_state WHERE room_id=$1", room_id
        )
        if not val:
            return None
        return RoomEncryptionStateEventContent.parse_json(val)



from .base import BaseConfig, BaseMissingError, ConfigUpdateHelper
from .file import BaseFileConfig, yaml
from .proxy import BaseProxyConfig
from .recursive_dict import RecursiveDict
from .string import BaseStringConfig
from .validation import BaseValidatableConfig, ConfigValueError, ForbiddenDefault, ForbiddenKey

__all__ = [
    "BaseConfig",
    "BaseMissingError",
    "ConfigUpdateHelper",
    "BaseFileConfig",
    "yaml",
    "BaseProxyConfig",
    "RecursiveDict",
    "BaseStringConfig",
    "BaseValidatableConfig",
    "ConfigValueError",
    "ForbiddenDefault",
    "ForbiddenKey",
]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, NamedTuple
from abc import ABC, abstractmethod
from collections import defaultdict
from string import Template
import asyncio
import html
import logging
import time

from mautrix.appservice import AppService, IntentAPI
from mautrix.errors import MatrixError, MatrixRequestError, MForbidden, MNotFound
from mautrix.types import (
    JSON,
    EncryptionAlgorithm,
    EventID,
    EventType,
    Format,
    MessageEventContent,
    MessageType,
    RoomEncryptionStateEventContent,
    RoomID,
    RoomTombstoneStateEventContent,
    TextMessageEventContent,
    UserID,
)
from mautrix.util.logging import TraceLogger
from mautrix.util.simple_lock import SimpleLock

from .. import bridge as br


class RelaySender(NamedTuple):
    sender: br.BaseUser | None
    is_relay: bool


class RejectMatrixInvite(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class IgnoreMatrixInvite(Exception):
    pass


class DMCreateError(RejectMatrixInvite):
    """
    An error raised by :meth:`BasePortal.prepare_dm` if the DM can't be set up.

    The message in the exception will be sent to the user as a message before the ghost leaves.
    """


class BasePortal(ABC):
    log: TraceLogger = logging.getLogger("mau.portal")
    _async_get_locks: dict[Any, asyncio.Lock] = defaultdict(lambda: asyncio.Lock())
    disappearing_msg_class: type[br.AbstractDisappearingMessage] | None = None
    _disappearing_lock: asyncio.Lock | None
    az: AppService
    matrix: br.BaseMatrixHandler
    bridge: br.Bridge
    loop: asyncio.AbstractEventLoop
    main_intent: IntentAPI
    mxid: RoomID | None
    name: str | None
    encrypted: bool
    is_direct: bool
    backfill_lock: SimpleLock

    relay_user_id: UserID | None
    _relay_user: br.BaseUser | None
    relay_emote_to_text: bool = True
    relay_formatted_body: bool = True

    def __init__(self) -> None:
        self._disappearing_lock = asyncio.Lock() if self.disappearing_msg_class else None

    @abstractmethod
    async def save(self) -> None:
        pass

    @abstractmethod
    async def get_dm_puppet(self) -> br.BasePuppet | None:
        """
        Get the ghost representing the other end of this direct chat.

        Returns:
            A puppet entity, or ``None`` if this is not a 1:1 chat.
        """

    @abstractmethod
    async def handle_matrix_message(
        self, sender: br.BaseUser, message: MessageEventContent, event_id: EventID
    ) -> None:
        pass

    async def prepare_remote_dm(
        self, room_id: RoomID, invited_by: br.BaseUser, puppet: br.BasePuppet
    ) -> str:
        """
        Do whatever is needed on the remote platform to set up a direct chat between the user
        and the ghost. By default, this does nothing (and lets :meth:`setup_matrix_dm` handle
        everything).

        Args:
            room_id: The room ID that will be used.
            invited_by: The Matrix user who invited the ghost.
            puppet: The ghost who was invited.

        Returns:
            A simple message indicating what was done (will be sent as a notice to the room).
            If empty, the message won't be sent.

        Raises:
            DMCreateError: if the DM could not be created and the ghost should leave the room.
        """
        return "Portal to private chat created."

    async def postprocess_matrix_dm(self, user: br.BaseUser, puppet: br.BasePuppet) -> None:
        await self.update_bridge_info()

    async def reject_duplicate_dm(
        self, room_id: RoomID, invited_by: br.BaseUser, puppet: br.BasePuppet
    ) -> None:
        try:
            await puppet.default_mxid_intent.send_notice(
                room_id,
                text=f"You already have a private chat with me: {self.mxid}",
                html=(
                    "You already have a private chat with me: "
                    f"<a href='https://matrix.to/#/{self.mxid}'>Link to room</a>"
                ),
            )
        except Exception as e:
            self.log.debug(f"Failed to send notice to duplicate private chat room: {e}")

        try:
            await puppet.default_mxid_intent.send_state_event(
                room_id,
                event_type=EventType.ROOM_TOMBSTONE,
                content=RoomTombstoneStateEventContent(
                    replacement_room=self.mxid,
                    body="You already have a private chat with me",
                ),
            )
        except Exception as e:
            self.log.debug(f"Failed to send tombstone to duplicate private chat room: {e}")

        await puppet.default_mxid_intent.leave_room(room_id)

    async def accept_matrix_dm(
        self, room_id: RoomID, invited_by: br.BaseUser, puppet: br.BasePuppet
    ) -> None:
        """
        Set up a room as a direct chat portal.

        The ghost has already accepted the invite at this point, so this method needs to make it
        leave if the DM can't be created for some reason.

        By default, this checks if there's an existing portal and redirects the user there if it
        does exist. If a portal doesn't exist, this will call :meth:`prepare_matrix_dm` and then
        save the room ID, enable encryption and update bridge info. If the portal exists, but isn't
        usable, the old room will be cleaned up and the function will continue.

        Args:
            room_id: The room ID that will be used.
            invited_by: The Matrix user who invited the ghost.
            puppet: The ghost who was invited.
        """
        if self.mxid:
            try:
                portal_members = await self.main_intent.get_room_members(self.mxid)
            except (MForbidden, MNotFound):
                portal_members = []
            if invited_by.mxid in portal_members:
                await self.reject_duplicate_dm(room_id, invited_by, puppet)
                return
            self.log.debug(
                f"{invited_by.mxid} isn't in old portal room {self.mxid},"
                " cleaning up and accepting new room as the DM portal"
            )
            await self.cleanup_portal(
                message="User seems to have left DM portal", puppets_only=True
            )
        try:
            message = await self.prepare_remote_dm(room_id, invited_by, puppet)
        except DMCreateError as e:
            if e.message:
                await puppet.default_mxid_intent.send_notice(room_id, text=e.message)
            await puppet.default_mxid_intent.leave_room(room_id, reason="Failed to create DM")
            return
        self.mxid = room_id
        e2be_ok = await self.check_dm_encryption()
        await self.save()
        if e2be_ok is False:
            message += "\n\nWarning: Failed to enable end-to-bridge encryption."
        if message:
            await self._send_message(
                puppet.default_mxid_intent,
                TextMessageEventContent(
                    msgtype=MessageType.NOTICE,
                    body=message,
                ),
            )
        await self.postprocess_matrix_dm(invited_by, puppet)

    async def handle_matrix_invite(self, invited_by: br.BaseUser, puppet: br.BasePuppet) -> None:
        """
        Called when a Matrix user invites a bridge ghost to a room to process the invite (and check
        if it should be accepted).

        Args:
            invited_by: The user who invited the ghost.
            puppet: The ghost who was invited.

        Raises:
            RejectMatrixInvite: if the invite should be rejected.
            IgnoreMatrixInvite: if the invite should be ignored (e.g. if it was already accepted).
        """
        if self.is_direct:
            raise RejectMatrixInvite("You can't invite additional users to private chats.")
        raise RejectMatrixInvite("This bridge does not implement inviting users to portals.")

    async def update_bridge_info(self) -> None:
        """Resend the ``m.bridge`` event into the room."""

    @property
    def _relay_is_implemented(self) -> bool:
        return hasattr(self, "relay_user_id") and hasattr(self, "_relay_user")

    @property
    def has_relay(self) -> bool:
        return (
            self._relay_is_implemented
            and self.bridge.config["bridge.relay.enabled"]
            and bool(self.relay_user_id)
        )

    async def get_relay_user(self) -> br.BaseUser | None:
        if not self.has_relay:
            return None
        if self._relay_user is None:
            self._relay_user = await self.bridge.get_user(self.relay_user_id)
        return self._relay_user if await self._relay_user.is_logged_in() else None

    async def set_relay_user(self, user: br.BaseUser | None) -> None:
        if not self._relay_is_implemented or not self.bridge.config["bridge.relay.enabled"]:
            raise RuntimeError("Can't set_relay_user() when relay mode is not enabled")
        self._relay_user = user
        self.relay_user_id = user.mxid if user else None
        await self.save()

    async def get_relay_sender(self, sender: br.BaseUser, evt_identifier: str) -> RelaySender:
        if not await sender.needs_relay(self):
            return RelaySender(sender, False)

        if not self.has_relay:
            self.log.debug(
                f"Ignoring {evt_identifier} from non-logged-in user {sender.mxid} "
                f"in chat with no relay user"
            )
            return RelaySender(None, True)
        relay_sender = await self.get_relay_user()
        if not relay_sender:
            self.log.debug(
                f"Ignoring {evt_identifier} from non-logged-in user {sender.mxid} "
                f"relay user {self.relay_user_id} is not set up correctly"
            )
            return RelaySender(None, True)
        return RelaySender(relay_sender, True)

    async def apply_relay_message_format(
        self, sender: br.BaseUser, content: MessageEventContent
    ) -> None:
        if self.relay_formatted_body and content.get("format", None) != Format.HTML:
            content["format"] = Format.HTML
            content["formatted_body"] = html.escape(content.body).replace("\n", "<br/>")
        tpl = self.bridge.config["bridge.relay.message_formats"].get(
            content.msgtype.value, "$sender_displayname: $message"
        )
        displayname = await self.get_displayname(sender)
        username, _ = self.az.intent.parse_user_id(sender.mxid)
        tpl_args = {
            "sender_mxid": sender.mxid,
            "sender_username": username,
            "sender_displayname": html.escape(displayname),
            "formatted_body": content["formatted_body"],
            "body": content.body,
            "message": content.body,
        }
        content.body = Template(tpl).safe_substitute(tpl_args)
        if self.relay_formatted_body and "formatted_body" in content:
            tpl_args["message"] = content["formatted_body"]
            content["formatted_body"] = Template(tpl).safe_substitute(tpl_args)
        if self.relay_emote_to_text and content.msgtype == MessageType.EMOTE:
            content.msgtype = MessageType.TEXT

    async def get_displayname(self, user: br.BaseUser) -> str:
        return await self.main_intent.get_room_displayname(self.mxid, user.mxid) or user.mxid

    async def check_dm_encryption(self) -> bool | None:
        try:
            evt = await self.main_intent.get_state_event(self.mxid, EventType.ROOM_ENCRYPTION)
            self.log.debug("Found existing encryption event in direct portal: %s", evt)
            if evt and evt.algorithm == EncryptionAlgorithm.MEGOLM_V1:
                self.encrypted = True
        except MNotFound:
            pass
        if (
            self.is_direct
            and self.matrix.e2ee
            and (self.bridge.config["bridge.encryption.default"] or self.encrypted)
        ):
            return await self.enable_dm_encryption()
        return None

    def get_encryption_state_event_json(self) -> JSON:
        evt = RoomEncryptionStateEventContent(EncryptionAlgorithm.MEGOLM_V1)
        if self.bridge.config["bridge.encryption.rotation.enable_custom"]:
            evt.rotation_period_ms = self.bridge.config["bridge.encryption.rotation.milliseconds"]
            evt.rotation_period_msgs = self.bridge.config["bridge.encryption.rotation.messages"]
        return evt.serialize()

    async def enable_dm_encryption(self) -> bool:
        self.log.debug("Inviting bridge bot to room for end-to-bridge encryption")
        try:
            await self.main_intent.invite_user(self.mxid, self.az.bot_mxid)
            await self.az.intent.join_room_by_id(self.mxid)
            if not self.encrypted:
                await self.main_intent.send_state_event(
                    self.mxid,
                    EventType.ROOM_ENCRYPTION,
                    self.get_encryption_state_event_json(),
                )
        except Exception:
            self.log.warning(f"Failed to enable end-to-bridge encryption", exc_info=True)
            return False

        self.encrypted = True
        await self.update_info_from_puppet()
        return True

    async def update_info_from_puppet(self, puppet: br.BasePuppet | None = None) -> None:
        """
        Update the room metadata to match the ghost's name/avatar.

        This is called after enabling encryption, as the bridge bot needs to join for e2ee,
        but that messes up the default name generation. If/when canonical DMs happen,
        this might not be necessary anymore.

        Args:
            puppet: The ghost that is the other participant in the room.
                If ``None``, the entity should be fetched as necessary.
        """

    @property
    def disappearing_enabled(self) -> bool:
        return bool(self.disappearing_msg_class)

    async def _disappear_event(self, msg: br.AbstractDisappearingMessage) -> None:
        sleep_time = (msg.expiration_ts / 1000) - time.time()
        self.log.trace(f"Sleeping {sleep_time:.3f} seconds before redacting {msg.event_id}")
        await asyncio.sleep(sleep_time)
        try:
            await msg.delete()
        except Exception:
            self.log.exception(
                f"Failed to delete disappearing message record for {msg.event_id} from database"
            )
        if self.mxid != msg.room_id:
            self.log.debug(
                f"Not redacting expired event {msg.event_id}, "
                f"portal room seems to have changed ({self.mxid!r} != {msg.room_id!r})"
            )
            return
        try:
            await self._do_disappear(msg.event_id)
            self.log.debug(f"Expired event {msg.event_id} disappeared successfully")
        except Exception as e:
            self.log.warning(f"Failed to make expired event {msg.event_id} disappear: {e}", e)

    async def _do_disappear(self, event_id: EventID) -> None:
        await self.main_intent.redact(self.mxid, event_id)

    @classmethod
    async def restart_scheduled_disappearing(cls) -> None:
        """
        Restart disappearing message timers for all messages that were already scheduled to
        disappear earlier. This should be called at bridge startup.
        """
        if not cls.disappearing_msg_class:
            return
        msgs = await cls.disappearing_msg_class.get_all_scheduled()
        for msg in msgs:
            portal = await cls.bridge.get_portal(msg.room_id)
            if portal and portal.mxid:
                asyncio.create_task(portal._disappear_event(msg))
            else:
                await msg.delete()

    async def schedule_disappearing(self) -> None:
        """
        Start the disappearing message timer for all unscheduled messages in this room.
        This is automatically called from :meth:`MatrixHandler.handle_receipt`.
        """
        if not self.disappearing_msg_class:
            return
        async with self._disappearing_lock:
            msgs = await self.disappearing_msg_class.get_unscheduled_for_room(self.mxid)
            for msg in msgs:
                msg.start_timer()
                await msg.update()
                asyncio.create_task(self._disappear_event(msg))

    async def _send_message(
        self,
        intent: IntentAPI,
        content: MessageEventContent,
        event_type: EventType = EventType.ROOM_MESSAGE,
        **kwargs,
    ) -> EventID:
        if self.encrypted and self.matrix.e2ee:
            event_type, content = await self.matrix.e2ee.encrypt(self.mxid, event_type, content)
        event_id = await intent.send_message_event(self.mxid, event_type, content, **kwargs)
        if intent.api.is_real_user:
            asyncio.create_task(intent.mark_read(self.mxid, event_id))
        return event_id

    @property
    @abstractmethod
    def bridge_info_state_key(self) -> str:
        pass

    @property
    @abstractmethod
    def bridge_info(self) -> dict[str, Any]:
        pass

    # region Matrix room cleanup

    @abstractmethod
    async def delete(self) -> None:
        pass

    @classmethod
    async def cleanup_room(
        cls,
        intent: IntentAPI,
        room_id: RoomID,
        message: str = "Cleaning room",
        puppets_only: bool = False,
    ) -> None:
        try:
            members = await intent.get_room_members(room_id)
        except MatrixError:
            members = []
        for user_id in members:
            if user_id == intent.mxid:
                continue

            puppet = await cls.bridge.get_puppet(user_id, create=False)
            if puppet:
                await puppet.default_mxid_intent.leave_room(room_id)
                continue

            if not puppets_only:
                custom_puppet = await cls.bridge.get_double_puppet(user_id)
                left = False
                if custom_puppet:
                    try:
                        await custom_puppet.intent.leave_room(room_id)
                        await custom_puppet.intent.forget_room(room_id)
                    except MatrixError:
                        pass
                    else:
                        left = True
                if not left:
                    try:
                        await intent.kick_user(room_id, user_id, message)
                    except MatrixError:
                        pass
        try:
            await intent.leave_room(room_id)
        except MatrixError:
            cls.log.warning(f"Failed to leave room {room_id} when cleaning up room", exc_info=True)

    async def cleanup_portal(self, message: str, puppets_only: bool = False) -> None:
        await self.cleanup_room(self.main_intent, self.mxid, message, puppets_only)
        await self.delete()

    async def unbridge(self) -> None:
        await self.cleanup_portal("Room unbridged", puppets_only=True)

    async def cleanup_and_delete(self) -> None:
        await self.cleanup_portal("Portal deleted")

    async def get_authenticated_matrix_users(self) -> list[UserID]:
        """
        Get the list of Matrix user IDs who can be bridged. This is used to determine if the portal
        is empty (and should be cleaned up) or not. Bridges should override this to check that the
        users are either logged in or the portal has a relaybot.
        """
        try:
            members = await self.main_intent.get_room_members(self.mxid)
        except MatrixRequestError:
            return []
        return [
            member
            for member in members
            if (not self.bridge.is_bridge_ghost(member) and member != self.az.bot_mxid)
        ]

    # endregion



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from mautrix.client.state_store.asyncpg import PgStateStore as PgClientStateStore
from mautrix.util.async_db import Database

from .memory import ASStateStore


class PgASStateStore(PgClientStateStore, ASStateStore):
    def __init__(self, db: Database) -> None:
        PgClientStateStore.__init__(self, db)
        ASStateStore.__init__(self)



from .auth import (
    DiscoveryInformation,
    DiscoveryIntegrations,
    DiscoveryIntegrationServer,
    DiscoveryServer,
    LoginFlow,
    LoginFlowList,
    LoginResponse,
    LoginType,
    MatrixUserIdentifier,
    PhoneIdentifier,
    ThirdPartyIdentifier,
    UserIdentifier,
    UserIdentifierType,
    WhoamiResponse,
)
from .crypto import (
    ClaimKeysResponse,
    CrossSigner,
    CrossSigningKeys,
    CrossSigningUsage,
    DecryptedOlmEvent,
    DeviceIdentity,
    DeviceKeys,
    OlmEventKeys,
    QueryKeysResponse,
    TOFUSigningKey,
    TrustState,
    UnsignedDeviceInfo,
)
from .event import (
    AccountDataEvent,
    AccountDataEventContent,
    ASToDeviceEvent,
    AudioInfo,
    BaseEvent,
    BaseFileInfo,
    BaseMessageEventContent,
    BaseMessageEventContentFuncs,
    BaseRoomEvent,
    BaseUnsigned,
    BatchSendEvent,
    BatchSendStateEvent,
    BeeperMessageStatusEvent,
    BeeperMessageStatusEventContent,
    CallAnswerEventContent,
    CallCandidate,
    CallCandidatesEventContent,
    CallData,
    CallDataType,
    CallEvent,
    CallEventContent,
    CallHangupEventContent,
    CallHangupReason,
    CallInviteEventContent,
    CallNegotiateEventContent,
    CallRejectEventContent,
    CallSelectAnswerEventContent,
    CanonicalAliasStateEventContent,
    EncryptedEvent,
    EncryptedEventContent,
    EncryptedFile,
    EncryptedMegolmEventContent,
    EncryptedOlmEventContent,
    EncryptionAlgorithm,
    EncryptionKeyAlgorithm,
    EphemeralEvent,
    Event,
    EventContent,
    EventType,
    FileInfo,
    Format,
    ForwardedRoomKeyEventContent,
    GenericEvent,
    ImageInfo,
    InReplyTo,
    JoinRule,
    JoinRulesStateEventContent,
    JSONWebKey,
    KeyID,
    KeyRequestAction,
    LocationInfo,
    LocationMessageEventContent,
    MediaInfo,
    MediaMessageEventContent,
    Membership,
    MemberStateEventContent,
    MessageEvent,
    MessageEventContent,
    MessageStatus,
    MessageStatusReason,
    MessageType,
    MessageUnsigned,
    OlmCiphertext,
    OlmMsgType,
    PowerLevelStateEventContent,
    PresenceEvent,
    PresenceEventContent,
    PresenceState,
    ReactionEvent,
    ReactionEventContent,
    ReceiptEvent,
    ReceiptEventContent,
    ReceiptType,
    RedactionEvent,
    RedactionEventContent,
    RelatesTo,
    RelationType,
    RequestedKeyInfo,
    RoomAvatarStateEventContent,
    RoomCreateStateEventContent,
    RoomEncryptionStateEventContent,
    RoomKeyEventContent,
    RoomKeyRequestEventContent,
    RoomKeyWithheldCode,
    RoomKeyWithheldEventContent,
    RoomNameStateEventContent,
    RoomPinnedEventsStateEventContent,
    RoomPredecessor,
    RoomTagAccountDataEventContent,
    RoomTagInfo,
    RoomTombstoneStateEventContent,
    RoomTopicStateEventContent,
    RoomType,
    SingleReceiptEventContent,
    SpaceChildStateEventContent,
    SpaceParentStateEventContent,
    StateEvent,
    StateEventContent,
    StateUnsigned,
    StrippedStateEvent,
    TextMessageEventContent,
    ThumbnailInfo,
    ToDeviceEvent,
    ToDeviceEventContent,
    TypingEvent,
    TypingEventContent,
    VideoInfo,
)
from .filter import EventFilter, Filter, RoomEventFilter, RoomFilter, StateFilter
from .matrixuri import IdentifierType, MatrixURI, MatrixURIError, URIAction
from .media import (
    MediaCreateResponse,
    MediaRepoConfig,
    MXOpenGraph,
    OpenGraphAudio,
    OpenGraphImage,
    OpenGraphVideo,
)
from .misc import (
    BatchSendResponse,
    DeviceLists,
    DeviceOTKCount,
    DirectoryPaginationToken,
    EventContext,
    PaginatedMessages,
    PaginationDirection,
    RoomAliasInfo,
    RoomCreatePreset,
    RoomDirectoryResponse,
    RoomDirectoryVisibility,
)
from .primitive import (
    JSON,
    BatchID,
    ContentURI,
    DeviceID,
    EventID,
    FilterID,
    IdentityKey,
    RoomAlias,
    RoomID,
    SessionID,
    Signature,
    SigningKey,
    SyncToken,
    UserID,
)
from .push_rules import (
    PushAction,
    PushActionDict,
    PushActionType,
    PushCondition,
    PushConditionKind,
    PushOperator,
    PushRule,
    PushRuleID,
    PushRuleKind,
    PushRuleScope,
)
from .users import Member, User, UserSearchResults
from .util import (
    ExtensibleEnum,
    Lst,
    Obj,
    Serializable,
    SerializableAttrs,
    SerializableEnum,
    SerializerError,
    deserializer,
    field,
    serializer,
)
from .versions import SpecVersions, Version, VersionFormat, VersionsResponse

__all__ = [
    "DiscoveryInformation",
    "DiscoveryIntegrations",
    "DiscoveryIntegrationServer",
    "DiscoveryServer",
    "LoginFlow",
    "LoginFlowList",
    "LoginResponse",
    "LoginType",
    "MatrixUserIdentifier",
    "PhoneIdentifier",
    "ThirdPartyIdentifier",
    "UserIdentifier",
    "UserIdentifierType",
    "WhoamiResponse",
    "ClaimKeysResponse",
    "CrossSigner",
    "CrossSigningKeys",
    "CrossSigningUsage",
    "DecryptedOlmEvent",
    "DeviceIdentity",
    "DeviceKeys",
    "OlmEventKeys",
    "QueryKeysResponse",
    "TOFUSigningKey",
    "TrustState",
    "UnsignedDeviceInfo",
    "AccountDataEvent",
    "AccountDataEventContent",
    "ASToDeviceEvent",
    "AudioInfo",
    "BaseEvent",
    "BaseFileInfo",
    "BaseMessageEventContent",
    "BaseMessageEventContentFuncs",
    "BaseRoomEvent",
    "BaseUnsigned",
    "BatchSendEvent",
    "BatchSendStateEvent",
    "BeeperMessageStatusEvent",
    "BeeperMessageStatusEventContent",
    "CallAnswerEventContent",
    "CallCandidate",
    "CallCandidatesEventContent",
    "CallData",
    "CallDataType",
    "CallEvent",
    "CallEventContent",
    "CallHangupEventContent",
    "CallHangupReason",
    "CallInviteEventContent",
    "CallNegotiateEventContent",
    "CallRejectEventContent",
    "CallSelectAnswerEventContent",
    "CanonicalAliasStateEventContent",
    "EncryptedEvent",
    "EncryptedEventContent",
    "EncryptedFile",
    "EncryptedMegolmEventContent",
    "EncryptedOlmEventContent",
    "EncryptionAlgorithm",
    "EncryptionKeyAlgorithm",
    "EphemeralEvent",
    "Event",
    "EventContent",
    "EventType",
    "FileInfo",
    "Format",
    "ForwardedRoomKeyEventContent",
    "GenericEvent",
    "ImageInfo",
    "InReplyTo",
    "JoinRule",
    "JoinRulesStateEventContent",
    "JSONWebKey",
    "KeyID",
    "KeyRequestAction",
    "LocationInfo",
    "LocationMessageEventContent",
    "MediaInfo",
    "MediaMessageEventContent",
    "Membership",
    "MemberStateEventContent",
    "MessageEvent",
    "MessageEventContent",
    "MessageStatus",
    "MessageStatusReason",
    "MessageType",
    "MessageUnsigned",
    "OlmCiphertext",
    "OlmMsgType",
    "PowerLevelStateEventContent",
    "PresenceEvent",
    "PresenceEventContent",
    "PresenceState",
    "ReactionEvent",
    "ReactionEventContent",
    "ReceiptEvent",
    "ReceiptEventContent",
    "ReceiptType",
    "RedactionEvent",
    "RedactionEventContent",
    "RelatesTo",
    "RelationType",
    "RequestedKeyInfo",
    "RoomAvatarStateEventContent",
    "RoomCreateStateEventContent",
    "RoomEncryptionStateEventContent",
    "RoomKeyEventContent",
    "RoomKeyRequestEventContent",
    "RoomKeyWithheldCode",
    "RoomKeyWithheldEventContent",
    "RoomNameStateEventContent",
    "RoomPinnedEventsStateEventContent",
    "RoomPredecessor",
    "RoomTagAccountDataEventContent",
    "RoomTagInfo",
    "RoomTombstoneStateEventContent",
    "RoomTopicStateEventContent",
    "RoomType",
    "SingleReceiptEventContent",
    "SpaceChildStateEventContent",
    "SpaceParentStateEventContent",
    "StateEvent",
    "StateEventContent",
    "StateUnsigned",
    "StrippedStateEvent",
    "TextMessageEventContent",
    "ThumbnailInfo",
    "ToDeviceEvent",
    "ToDeviceEventContent",
    "TypingEvent",
    "TypingEventContent",
    "VideoInfo",
    "EventFilter",
    "Filter",
    "RoomEventFilter",
    "RoomFilter",
    "StateFilter",
    "IdentifierType",
    "MatrixURI",
    "MatrixURIError",
    "URIAction",
    "MediaCreateResponse",
    "MediaRepoConfig",
    "MXOpenGraph",
    "OpenGraphAudio",
    "OpenGraphImage",
    "OpenGraphVideo",
    "BatchSendResponse",
    "DeviceLists",
    "DeviceOTKCount",
    "DirectoryPaginationToken",
    "EventContext",
    "PaginatedMessages",
    "PaginationDirection",
    "RoomAliasInfo",
    "RoomCreatePreset",
    "RoomDirectoryResponse",
    "RoomDirectoryVisibility",
    "JSON",
    "BatchID",
    "ContentURI",
    "DeviceID",
    "EventID",
    "FilterID",
    "IdentityKey",
    "RoomAlias",
    "RoomID",
    "SessionID",
    "Signature",
    "SigningKey",
    "SyncToken",
    "UserID",
    "PushAction",
    "PushActionDict",
    "PushActionType",
    "PushCondition",
    "PushConditionKind",
    "PushOperator",
    "PushRule",
    "PushRuleID",
    "PushRuleKind",
    "PushRuleScope",
    "Member",
    "User",
    "UserSearchResults",
    "ExtensibleEnum",
    "Lst",
    "Obj",
    "Serializable",
    "SerializableAttrs",
    "SerializableEnum",
    "SerializerError",
    "deserializer",
    "field",
    "serializer",
    "SpecVersions",
    "Version",
    "VersionFormat",
    "VersionsResponse",
]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any
from contextlib import asynccontextmanager
import asyncio
import logging
import sys
import traceback

from yarl import URL
import asyncpg

from .connection import LoggingConnection
from .database import Database
from .scheme import Scheme
from .upgrade import UpgradeTable


class PostgresDatabase(Database):
    scheme = Scheme.POSTGRES
    _pool: asyncpg.pool.Pool | None
    _pool_override: bool
    _exit_on_ice: bool

    def __init__(
        self,
        url: URL,
        upgrade_table: UpgradeTable,
        db_args: dict[str, Any] = None,
        log: logging.Logger | None = None,
        owner_name: str | None = None,
        ignore_foreign_tables: bool = True,
    ) -> None:
        if url.scheme in ("cockroach", "cockroachdb"):
            self.scheme = Scheme.COCKROACH
            # Send postgres scheme to asyncpg
            url = url.with_scheme("postgres")
        self._exit_on_ice = True
        if db_args:
            self._exit_on_ice = db_args.pop("meow_exit_on_ice", True)
            db_args.pop("init_commands", None)
        super().__init__(
            url,
            db_args=db_args,
            upgrade_table=upgrade_table,
            log=log,
            owner_name=owner_name,
            ignore_foreign_tables=ignore_foreign_tables,
        )
        self._pool = None
        self._pool_override = False

    def override_pool(self, db: PostgresDatabase) -> None:
        self._pool = db._pool
        self._pool_override = True

    async def start(self) -> None:
        if not self._pool_override:
            if self._pool:
                raise RuntimeError("Database has already been started")
            self._db_args["loop"] = asyncio.get_running_loop()
            log_url = self.url
            if log_url.password:
                log_url = log_url.with_password("password-redacted")
            self.log.debug(f"Connecting to {log_url}")
            self._pool = await asyncpg.create_pool(str(self.url), **self._db_args)
        await super().start()

    @property
    def pool(self) -> asyncpg.pool.Pool:
        if not self._pool:
            raise RuntimeError("Database has not been started")
        return self._pool

    async def stop(self) -> None:
        if not self._pool_override and self._pool is not None:
            await self._pool.close()

    async def _handle_exception(self, err: Exception) -> None:
        if self._exit_on_ice and isinstance(err, asyncpg.InternalClientError):
            pre_stack = traceback.format_stack()[:-2]
            post_stack = traceback.format_exception(err)
            header = post_stack[0]
            post_stack = post_stack[1:]
            self.log.critical(
                "Got asyncpg internal client error, exiting...\n%s%s%s",
                header,
                "".join(pre_stack),
                "".join(post_stack),
            )
            sys.exit(26)

    @asynccontextmanager
    async def acquire(self) -> LoggingConnection:
        async with self.pool.acquire() as conn:
            yield LoggingConnection(
                self.scheme, conn, self.log, handle_exception=self._handle_exception
            )


Database.schemes["postgres"] = PostgresDatabase
Database.schemes["postgresql"] = PostgresDatabase
Database.schemes["cockroach"] = PostgresDatabase
Database.schemes["cockroachdb"] = PostgresDatabase



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import ClassVar
from abc import ABC, abstractmethod

from mautrix.types import Event, EventType, Membership, StateEvent

from . import syncer


class Dispatcher(ABC):
    client: syncer.Syncer

    def __init__(self, client: syncer.Syncer) -> None:
        self.client = client

    @abstractmethod
    def register(self) -> None:
        pass

    @abstractmethod
    def unregister(self) -> None:
        pass


class SimpleDispatcher(Dispatcher, ABC):
    event_type: ClassVar[EventType]

    def register(self) -> None:
        self.client.add_event_handler(self.event_type, self.handle)

    def unregister(self) -> None:
        self.client.remove_event_handler(self.event_type, self.handle)

    @abstractmethod
    async def handle(self, evt: Event) -> None:
        pass


class MembershipEventDispatcher(SimpleDispatcher):
    event_type = EventType.ROOM_MEMBER

    async def handle(self, evt: StateEvent) -> None:
        if evt.type != EventType.ROOM_MEMBER:
            return

        if evt.content.membership == Membership.JOIN:
            if evt.prev_content.membership != Membership.JOIN:
                change_type = syncer.InternalEventType.JOIN
            else:
                change_type = syncer.InternalEventType.PROFILE_CHANGE
        elif evt.content.membership == Membership.INVITE:
            change_type = syncer.InternalEventType.INVITE
        elif evt.content.membership == Membership.LEAVE:
            if evt.prev_content.membership == Membership.BAN:
                change_type = syncer.InternalEventType.UNBAN
            elif evt.prev_content.membership == Membership.INVITE:
                if evt.state_key == evt.sender:
                    change_type = syncer.InternalEventType.REJECT_INVITE
                else:
                    change_type = syncer.InternalEventType.DISINVITE
            elif evt.state_key == evt.sender:
                change_type = syncer.InternalEventType.LEAVE
            else:
                change_type = syncer.InternalEventType.KICK
        elif evt.content.membership == Membership.BAN:
            change_type = syncer.InternalEventType.BAN
        else:
            return

        self.client.dispatch_manual_event(change_type, evt)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, List, Optional, Pattern, Union
from html import escape
import re

from attr import dataclass
import attr

from ..primitive import JSON, ContentURI, EventID
from ..util import ExtensibleEnum, Obj, SerializableAttrs, deserializer, field
from .base import BaseRoomEvent, BaseUnsigned

# region Message types


class Format(ExtensibleEnum):
    """A message format. Currently only ``org.matrix.custom.html`` is available.
    This will probably be deprecated when extensible events are implemented."""

    HTML: "Format" = "org.matrix.custom.html"


TEXT_MESSAGE_TYPES = ("m.text", "m.emote", "m.notice")
MEDIA_MESSAGE_TYPES = ("m.image", "m.sticker", "m.video", "m.audio", "m.file")


class MessageType(ExtensibleEnum):
    """A message type."""

    TEXT: "MessageType" = "m.text"
    EMOTE: "MessageType" = "m.emote"
    NOTICE: "MessageType" = "m.notice"
    IMAGE: "MessageType" = "m.image"
    STICKER: "MessageType" = "m.sticker"
    VIDEO: "MessageType" = "m.video"
    AUDIO: "MessageType" = "m.audio"
    FILE: "MessageType" = "m.file"
    LOCATION: "MessageType" = "m.location"

    @property
    def is_text(self) -> bool:
        return self.value in TEXT_MESSAGE_TYPES

    @property
    def is_media(self) -> bool:
        return self.value in MEDIA_MESSAGE_TYPES


# endregion
# region Relations


@dataclass
class InReplyTo(SerializableAttrs):
    event_id: EventID


class RelationType(ExtensibleEnum):
    ANNOTATION: "RelationType" = "m.annotation"
    REFERENCE: "RelationType" = "m.reference"
    REPLACE: "RelationType" = "m.replace"
    THREAD: "RelationType" = "m.thread"


@dataclass
class RelatesTo(SerializableAttrs):
    """Message relations. Used for reactions, edits and replies."""

    rel_type: RelationType = None
    event_id: Optional[EventID] = None
    key: Optional[str] = None
    is_falling_back: Optional[bool] = None
    in_reply_to: Optional[InReplyTo] = field(default=None, json="m.in_reply_to")

    def __bool__(self) -> bool:
        return (bool(self.rel_type) and bool(self.event_id)) or bool(self.in_reply_to)

    def serialize(self) -> JSON:
        if not self:
            return attr.NOTHING
        return super().serialize()


# endregion
# region Base event content


class BaseMessageEventContentFuncs:
    """Base class for the contents of all message-type events (currently m.room.message and
    m.sticker). Contains relation helpers."""

    body: str
    _relates_to: Optional[RelatesTo]

    def set_reply(self, reply_to: Union[EventID, "MessageEvent"], **kwargs) -> None:
        self.relates_to.in_reply_to = InReplyTo(
            event_id=reply_to if isinstance(reply_to, str) else reply_to.event_id
        )

    def set_thread_parent(
        self,
        thread_parent: Union[EventID, "MessageEvent"],
        last_event_in_thread: Union[EventID, "MessageEvent", None] = None,
        disable_reply_fallback: bool = False,
        **kwargs,
    ) -> None:
        self.relates_to.rel_type = RelationType.THREAD
        self.relates_to.event_id = (
            thread_parent if isinstance(thread_parent, str) else thread_parent.event_id
        )
        if not disable_reply_fallback:
            self.set_reply(last_event_in_thread or thread_parent, **kwargs)
            self.relates_to.is_falling_back = True

    def set_edit(self, edits: Union[EventID, "MessageEvent"]) -> None:
        self.relates_to.rel_type = RelationType.REPLACE
        self.relates_to.event_id = edits if isinstance(edits, str) else edits.event_id
        # Library consumers may create message content by setting a reply first,
        # then later marking it as an edit. As edits can't change the reply, just remove
        # the reply metadata when marking as a reply.
        if self.relates_to.in_reply_to:
            self.relates_to.in_reply_to = None
            self.relates_to.is_falling_back = None

    def serialize(self) -> JSON:
        data = SerializableAttrs.serialize(self)
        evt = self.get_edit()
        if evt:
            new_content = {**data}
            del new_content["m.relates_to"]
            data["m.new_content"] = new_content
            if "body" in data:
                data["body"] = f"* {data['body']}"
            if "formatted_body" in data:
                data["formatted_body"] = f"* {data['formatted_body']}"
        return data

    @property
    def relates_to(self) -> RelatesTo:
        if self._relates_to is None:
            self._relates_to = RelatesTo()
        return self._relates_to

    @relates_to.setter
    def relates_to(self, relates_to: RelatesTo) -> None:
        self._relates_to = relates_to

    def get_reply_to(self) -> Optional[EventID]:
        if self._relates_to and self._relates_to.in_reply_to:
            return self._relates_to.in_reply_to.event_id
        return None

    def get_edit(self) -> Optional[EventID]:
        if self._relates_to and self._relates_to.rel_type == RelationType.REPLACE:
            return self._relates_to.event_id
        return None

    def get_thread_parent(self) -> Optional[EventID]:
        if self._relates_to and self._relates_to.rel_type == RelationType.THREAD:
            return self._relates_to.event_id
        return None

    def trim_reply_fallback(self) -> None:
        pass


@dataclass
class BaseMessageEventContent(BaseMessageEventContentFuncs):
    """Base event content for all m.room.message-type events."""

    msgtype: MessageType = None
    body: str = ""

    external_url: str = None
    _relates_to: Optional[RelatesTo] = attr.ib(default=None, metadata={"json": "m.relates_to"})


# endregion
# region Media info


@dataclass
class JSONWebKey(SerializableAttrs):
    key: str = attr.ib(metadata={"json": "k"})
    algorithm: str = attr.ib(default="A256CTR", metadata={"json": "alg"})
    extractable: bool = attr.ib(default=True, metadata={"json": "ext"})
    key_type: str = attr.ib(default="oct", metadata={"json": "kty"})
    key_ops: List[str] = attr.ib(factory=lambda: ["encrypt", "decrypt"])


@dataclass
class EncryptedFile(SerializableAttrs):
    key: JSONWebKey
    iv: str
    hashes: Dict[str, str]
    url: Optional[ContentURI] = None
    version: str = attr.ib(default="v2", metadata={"json": "v"})


@dataclass
class BaseFileInfo(SerializableAttrs):
    mimetype: str = None
    size: int = None


@dataclass
class ThumbnailInfo(BaseFileInfo, SerializableAttrs):
    """Information about the thumbnail for a document, video, image or location."""

    height: int = attr.ib(default=None, metadata={"json": "h"})
    width: int = attr.ib(default=None, metadata={"json": "w"})
    orientation: int = None


@dataclass
class FileInfo(BaseFileInfo, SerializableAttrs):
    """Information about a document message."""

    thumbnail_info: Optional[ThumbnailInfo] = None
    thumbnail_file: Optional[EncryptedFile] = None
    thumbnail_url: Optional[ContentURI] = None


@dataclass
class ImageInfo(FileInfo, SerializableAttrs):
    """Information about an image message."""

    height: int = attr.ib(default=None, metadata={"json": "h"})
    width: int = attr.ib(default=None, metadata={"json": "w"})
    orientation: int = None


@dataclass
class VideoInfo(ImageInfo, SerializableAttrs):
    """Information about a video message."""

    duration: int = None
    orientation: int = None


@dataclass
class AudioInfo(BaseFileInfo, SerializableAttrs):
    """Information about an audio message."""

    duration: int = None


MediaInfo = Union[ImageInfo, VideoInfo, AudioInfo, FileInfo, Obj]


@dataclass
class LocationInfo(SerializableAttrs):
    """Information about a location message."""

    thumbnail_url: Optional[ContentURI] = None
    thumbnail_info: Optional[ThumbnailInfo] = None
    thumbnail_file: Optional[EncryptedFile] = None


# endregion
# region Event content


@dataclass
class MediaMessageEventContent(BaseMessageEventContent, SerializableAttrs):
    """The content of a media message event (m.image, m.audio, m.video, m.file)"""

    url: Optional[ContentURI] = None
    info: Optional[MediaInfo] = None
    file: Optional[EncryptedFile] = None

    @staticmethod
    @deserializer(MediaInfo)
    @deserializer(Optional[MediaInfo])
    def deserialize_info(data: JSON) -> MediaInfo:
        if not isinstance(data, dict):
            return Obj()
        msgtype = data.pop("__mautrix_msgtype", None)
        if msgtype == "m.image" or msgtype == "m.sticker":
            return ImageInfo.deserialize(data)
        elif msgtype == "m.video":
            return VideoInfo.deserialize(data)
        elif msgtype == "m.audio":
            return AudioInfo.deserialize(data)
        elif msgtype == "m.file":
            return FileInfo.deserialize(data)
        else:
            return Obj(**data)


@dataclass
class LocationMessageEventContent(BaseMessageEventContent, SerializableAttrs):
    geo_uri: str = None
    info: LocationInfo = None


html_reply_fallback_regex: Pattern = re.compile(r"^<mx-reply>[\s\S]+</mx-reply>")


@dataclass
class TextMessageEventContent(BaseMessageEventContent, SerializableAttrs):
    """The content of a text message event (m.text, m.notice, m.emote)"""

    format: Format = None
    formatted_body: str = None

    def set_reply(
        self, reply_to: Union["MessageEvent", EventID], *, displayname: Optional[str] = None
    ) -> None:
        super().set_reply(reply_to)
        if isinstance(reply_to, str):
            return
        if isinstance(reply_to, MessageEvent):
            self.ensure_has_html()
            if isinstance(reply_to.content, TextMessageEventContent):
                reply_to.content.trim_reply_fallback()
            self.formatted_body = (
                reply_to.make_reply_fallback_html(displayname) + self.formatted_body
            )
            self.body = reply_to.make_reply_fallback_text(displayname) + self.body

    def ensure_has_html(self) -> None:
        if not self.formatted_body or self.format != Format.HTML:
            self.format = Format.HTML
            self.formatted_body = escape(self.body).replace("\n", "<br/>")

    def formatted(self, format: Format) -> Optional[str]:
        if self.format == format:
            return self.formatted_body
        return None

    def trim_reply_fallback(self) -> None:
        if self.get_reply_to() and not getattr(self, "__reply_fallback_trimmed", False):
            self._trim_reply_fallback_text()
            self._trim_reply_fallback_html()
            setattr(self, "__reply_fallback_trimmed", True)

    def _trim_reply_fallback_text(self) -> None:
        if not self.body.startswith("> ") or "\n" not in self.body:
            return
        lines = self.body.split("\n")
        while len(lines) > 0 and lines[0].startswith("> "):
            lines.pop(0)
        # Pop extra newline at end of fallback
        lines.pop(0)
        self.body = "\n".join(lines)

    def _trim_reply_fallback_html(self) -> None:
        if self.formatted_body and self.format == Format.HTML:
            self.formatted_body = html_reply_fallback_regex.sub("", self.formatted_body)


MessageEventContent = Union[
    TextMessageEventContent, MediaMessageEventContent, LocationMessageEventContent, Obj
]


# endregion


@dataclass
class MessageUnsigned(BaseUnsigned, SerializableAttrs):
    """Unsigned information sent with message events."""

    transaction_id: str = None


html_reply_fallback_format = (
    "<mx-reply><blockquote>"
    "<a href='https://matrix.to/#/{room_id}/{event_id}'>In reply to</a> "
    "<a href='https://matrix.to/#/{sender}'>{displayname}</a><br/>"
    "{content}"
    "</blockquote></mx-reply>"
)

media_reply_fallback_body_map = {
    MessageType.IMAGE: "an image",
    MessageType.STICKER: "a sticker",
    MessageType.AUDIO: "audio",
    MessageType.VIDEO: "a video",
    MessageType.FILE: "a file",
    MessageType.LOCATION: "a location",
}


@dataclass
class MessageEvent(BaseRoomEvent, SerializableAttrs):
    """An m.room.message event"""

    content: MessageEventContent
    unsigned: Optional[MessageUnsigned] = field(factory=lambda: MessageUnsigned())

    @staticmethod
    @deserializer(MessageEventContent)
    def deserialize_content(data: JSON) -> MessageEventContent:
        if not isinstance(data, dict):
            return Obj()
        rel = data.get("m.relates_to", None) or {}
        if rel.get("rel_type", None) == RelationType.REPLACE.value:
            data = data.get("m.new_content", data)
            data["m.relates_to"] = rel
        msgtype = data.get("msgtype", None)
        if msgtype in TEXT_MESSAGE_TYPES:
            return TextMessageEventContent.deserialize(data)
        elif msgtype in MEDIA_MESSAGE_TYPES:
            data.get("info", {})["__mautrix_msgtype"] = msgtype
            return MediaMessageEventContent.deserialize(data)
        elif msgtype == "m.location":
            return LocationMessageEventContent.deserialize(data)
        else:
            return Obj(**data)

    def make_reply_fallback_html(self, displayname: Optional[str] = None) -> str:
        """Generate the HTML fallback for messages replying to this event."""
        if self.content.msgtype.is_text:
            body = self.content.formatted_body or escape(self.content.body).replace("\n", "<br/>")
        else:
            sent_type = media_reply_fallback_body_map[self.content.msgtype] or "a message"
            body = f"sent {sent_type}"
        displayname = escape(displayname) if displayname else self.sender
        return html_reply_fallback_format.format(
            room_id=self.room_id,
            event_id=self.event_id,
            sender=self.sender,
            displayname=displayname,
            content=body,
        )

    def make_reply_fallback_text(self, displayname: Optional[str] = None) -> str:
        """Generate the plaintext fallback for messages replying to this event."""
        if self.content.msgtype.is_text:
            body = self.content.body
        else:
            try:
                body = media_reply_fallback_body_map[self.content.msgtype]
            except KeyError:
                body = "an unknown message type"
        lines = body.strip().split("\n")
        first_line, lines = lines[0], lines[1:]
        fallback_text = f"> <{displayname or self.sender}> {first_line}"
        for line in lines:
            fallback_text += f"\n> {line}"
        fallback_text += "\n\n"
        return fallback_text



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Based on https://github.com/nhoad/aiomanhole Copyright (c) 2014, Nathan Hoad
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union
from abc import ABC, abstractmethod
from io import BytesIO, StringIO
from socket import SOL_SOCKET
from types import CodeType
import ast
import asyncio
import codeop
import contextlib
import functools
import inspect
import logging
import os
import pwd
import struct
import sys
import traceback

try:
    from socket import SO_PEERCRED
except ImportError:
    SO_PEERCRED = None

log = logging.getLogger("mau.manhole")


TOP_LEVEL_AWAIT = sys.version_info >= (3, 8)
ASYNC_EVAL_WRAPPER: str = """
async def __eval_async_expr():
    try:
        pass
    finally:
        globals().update(locals())
"""


def compile_async(tree: ast.AST) -> CodeType:
    flags = 0
    if TOP_LEVEL_AWAIT:
        flags += ast.PyCF_ALLOW_TOP_LEVEL_AWAIT
        node_to_compile = tree
    else:
        insert_returns(tree.body)
        wrapper_node: ast.AST = ast.parse(ASYNC_EVAL_WRAPPER, "<async eval wrapper>", "single")
        method_stmt = wrapper_node.body[0]
        try_stmt = method_stmt.body[0]
        try_stmt.body = tree.body
        node_to_compile = wrapper_node
    return compile(node_to_compile, "<manhole input>", "single", flags=flags)


# From https://gist.github.com/nitros12/2c3c265813121492655bc95aa54da6b9
def insert_returns(body: List[ast.AST]) -> None:
    if isinstance(body[-1], ast.Expr):
        body[-1] = ast.Return(body[-1].value)
        ast.fix_missing_locations(body[-1])
    elif isinstance(body[-1], ast.If):
        insert_returns(body[-1].body)
        insert_returns(body[-1].orelse)
    elif isinstance(body[-1], (ast.With, ast.AsyncWith)):
        insert_returns(body[-1].body)


class StatefulCommandCompiler(codeop.CommandCompiler):
    """A command compiler that buffers input until a full command is available."""

    buf: BytesIO

    def __init__(self) -> None:
        super().__init__()
        self.compiler = functools.partial(
            compile, optimize=1, flags=ast.PyCF_ONLY_AST | codeop.PyCF_DONT_IMPLY_DEDENT
        )
        self.buf = BytesIO()

    def is_partial_command(self) -> bool:
        return bool(self.buf.getvalue())

    def __call__(self, source: bytes, **kwargs: Any) -> Optional[CodeType]:
        buf = self.buf
        if self.is_partial_command():
            buf.write(b"\n")
        buf.write(source)

        code = self.buf.getvalue().decode("utf-8")
        codeobj = super().__call__(code, **kwargs)

        if codeobj:
            self.reset()
            return compile_async(codeobj)
        return None

    def reset(self) -> None:
        self.buf.seek(0)
        self.buf.truncate(0)


class Interpreter(ABC):
    @abstractmethod
    def __init__(self, namespace: Dict[str, Any], banner: Union[bytes, str]) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    async def __call__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        pass


class AsyncInterpreter(Interpreter):
    """An interactive asynchronous interpreter."""

    reader: asyncio.StreamReader
    writer: asyncio.StreamWriter
    namespace: Dict[str, Any]
    banner: bytes
    compiler: StatefulCommandCompiler
    running: bool

    def __init__(self, namespace: Dict[str, Any], banner: Union[bytes, str]) -> None:
        super().__init__(namespace, banner)
        self.namespace = namespace
        self.banner = banner if isinstance(banner, bytes) else str(banner).encode("utf-8")
        self.compiler = StatefulCommandCompiler()

    async def send_exception(self) -> None:
        """When an exception has occurred, write the traceback to the user."""
        self.compiler.reset()

        exc = traceback.format_exc()
        self.writer.write(exc.encode("utf-8"))

        await self.writer.drain()

    async def execute(self, codeobj: CodeType) -> Tuple[Any, str]:
        with contextlib.redirect_stdout(StringIO()) as buf:
            if TOP_LEVEL_AWAIT:
                value = eval(codeobj, self.namespace)
                if codeobj.co_flags & inspect.CO_COROUTINE:
                    value = await value
            else:
                exec(codeobj, self.namespace)
                value = await eval("__eval_async_expr()", self.namespace)

        return value, buf.getvalue()

    async def handle_one_command(self) -> None:
        """Process a single command. May have many lines."""

        while True:
            await self.write_prompt()
            codeobj = await self.read_command()

            if codeobj is not None:
                await self.run_command(codeobj)
                return

    async def run_command(self, codeobj: CodeType) -> None:
        """Execute a compiled code object, and write the output back to the client."""
        try:
            value, stdout = await self.execute(codeobj)
        except Exception:
            await self.send_exception()
            return
        else:
            await self.send_output(value, stdout)

    async def write_prompt(self) -> None:
        writer = self.writer

        if self.compiler.is_partial_command():
            writer.write(b"... ")
        else:
            writer.write(b">>> ")

        await writer.drain()

    async def read_command(self) -> Optional[CodeType]:
        """Read a command from the user line by line.

        Returns a code object suitable for execution.
        """

        reader = self.reader

        line = await reader.readline()
        if line == b"":
            raise ConnectionResetError()

        try:
            # skip the newline to make CommandCompiler work as advertised
            codeobj = self.compiler(line.rstrip(b"\n"))
        except SyntaxError:
            await self.send_exception()
            return None

        return codeobj

    async def send_output(self, value: str, stdout: str) -> None:
        """Write the output or value of the expression back to user.

        >>> 5
        5
        >>> print('cash rules everything around me')
        cash rules everything around me
        """

        writer = self.writer

        if value is not None:
            writer.write(f"{value!r}\n".encode("utf-8"))

        if stdout:
            writer.write(stdout.encode("utf-8"))

        await writer.drain()

    def close(self) -> None:
        if self.running:
            self.writer.close()
            self.running = False

    async def __call__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        """Main entry point for an interpreter session with a single client."""
        self.reader = reader
        self.writer = writer
        self.running = True

        if self.banner:
            writer.write(self.banner)
            await writer.drain()

        while self.running:
            try:
                await self.handle_one_command()
            except ConnectionResetError:
                writer.close()
                self.running = False
                break
            except Exception:
                log.exception("Exception in manhole REPL")
                self.writer.write(traceback.format_exc())
                await self.writer.drain()


class InterpreterFactory:
    namespace: Dict[str, Any]
    banner: bytes
    interpreter_class: Type[Interpreter]
    clients: List[Interpreter]
    whitelist: Set[int]
    _conn_id: int

    def __init__(
        self,
        namespace: Dict[str, Any],
        banner: Union[bytes, str],
        interpreter_class: Type[Interpreter],
        whitelist: Set[int],
    ) -> None:
        self.namespace = namespace or {}
        self.banner = banner
        self.interpreter_class = interpreter_class
        self.clients = []
        self.whitelist = whitelist
        self._conn_id = 0

    @property
    def conn_id(self) -> int:
        self._conn_id += 1
        return self._conn_id

    async def __call__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        sock = writer.transport.get_extra_info("socket")
        # TODO support non-linux OSes
        # I think FreeBSD uses SCM_CREDS
        creds = sock.getsockopt(SOL_SOCKET, SO_PEERCRED, struct.calcsize("3i"))
        pid, uid, gid = struct.unpack("3i", creds)
        user_info = pwd.getpwuid(uid)
        username = f"{user_info.pw_name} ({uid})" if user_info and user_info.pw_name else uid
        if len(self.whitelist) > 0 and uid not in self.whitelist:
            writer.write(b"You are not whitelisted to use the manhole.")
            log.warning(f"Non-whitelisted user {username} tried to connect from PID {pid}")
            await writer.drain()
            writer.close()
            return

        namespace = {**self.namespace}
        interpreter = self.interpreter_class(namespace=namespace, banner=self.banner)
        namespace["exit"] = interpreter.close
        self.clients.append(interpreter)
        conn_id = self.conn_id

        log.info(f"Manhole connection OPENED: {conn_id} from PID {pid} by {username}")
        await asyncio.create_task(interpreter(reader, writer))
        log.info(f"Manhole connection CLOSED: {conn_id} from PID {pid} by {username}")
        self.clients.remove(interpreter)


async def start_manhole(
    path: str,
    banner: str = "",
    namespace: Optional[Dict[str, Any]] = None,
    loop: asyncio.AbstractEventLoop = None,
    whitelist: Set[int] = None,
) -> Tuple[asyncio.AbstractServer, Callable[[], None]]:
    """
    Starts a manhole server on a given UNIX address.

    Args:
        path: The path to create the UNIX socket at.
        banner: The banner to show when clients connect.
        namespace: The globals to provide to connected clients.
        loop: The asyncio event loop to use.
        whitelist: List of user IDs to allow connecting.
    """
    if not SO_PEERCRED:
        raise ValueError("SO_PEERCRED is not supported on this platform")
    factory = InterpreterFactory(
        namespace=namespace,
        banner=banner,
        interpreter_class=AsyncInterpreter,
        whitelist=whitelist,
    )
    server = await asyncio.start_unix_server(factory, path=path)
    os.chmod(path, 0o666)

    def stop():
        for client in factory.clients:
            client.close()
        server.close()

    return server, stop



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, Optional, Set, Tuple, cast
from datetime import datetime, timedelta

from _libolm import ffi, lib
import olm

from mautrix.errors import EncryptionError
from mautrix.types import (
    DeviceID,
    EncryptionAlgorithm,
    IdentityKey,
    OlmCiphertext,
    OlmMsgType,
    RoomID,
    RoomKeyEventContent,
    SigningKey,
    UserID,
)


class Session(olm.Session):
    creation_time: datetime
    last_encrypted: datetime
    last_decrypted: datetime

    def __init__(self):
        super().__init__()
        self.creation_time = datetime.now()
        self.last_encrypted = datetime.now()
        self.last_decrypted = datetime.now()

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    @property
    def expired(self):
        return False

    @classmethod
    def from_pickle(
        cls,
        pickle: bytes,
        passphrase: str,
        creation_time: datetime,
        last_encrypted: Optional[datetime] = None,
        last_decrypted: Optional[datetime] = None,
    ) -> "Session":
        session = super().from_pickle(pickle, passphrase=passphrase)
        session.creation_time = creation_time
        session.last_encrypted = last_encrypted or creation_time
        session.last_decrypted = last_decrypted or creation_time
        return session

    def matches(self, ciphertext: str) -> bool:
        return super().matches(olm.OlmPreKeyMessage(ciphertext))

    def decrypt(self, ciphertext: OlmCiphertext) -> str:
        plaintext = super().decrypt(
            olm.OlmPreKeyMessage(ciphertext.body)
            if ciphertext.type == OlmMsgType.PREKEY
            else olm.OlmMessage(ciphertext.body)
        )
        self.last_decrypted = datetime.now()
        return plaintext

    def encrypt(self, plaintext: str) -> OlmCiphertext:
        self.last_encrypted = datetime.now()
        result = super().encrypt(plaintext)
        return OlmCiphertext(
            type=(
                OlmMsgType.PREKEY
                if isinstance(result, olm.OlmPreKeyMessage)
                else OlmMsgType.MESSAGE
            ),
            body=result.ciphertext,
        )

    def describe(self) -> str:
        parent = super()
        if hasattr(parent, "describe"):
            return parent.describe()
        elif hasattr(lib, "olm_session_describe"):
            describe_length = 600
            describe_buffer = ffi.new("char[]", describe_length)
            lib.olm_session_describe(self._session, describe_buffer, describe_length)
            return ffi.string(describe_buffer).decode("utf-8")
        else:
            return "describe not supported"


class InboundGroupSession(olm.InboundGroupSession):
    room_id: RoomID
    signing_key: SigningKey
    sender_key: IdentityKey
    forwarding_chain: List[IdentityKey]

    def __init__(
        self,
        session_key: str,
        signing_key: SigningKey,
        sender_key: IdentityKey,
        room_id: RoomID,
        forwarding_chain: Optional[List[IdentityKey]] = None,
    ) -> None:
        self.signing_key = signing_key
        self.sender_key = sender_key
        self.room_id = room_id
        self.forwarding_chain = forwarding_chain or []
        super().__init__(session_key)

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    @classmethod
    def from_pickle(
        cls,
        pickle: bytes,
        passphrase: str,
        signing_key: SigningKey,
        sender_key: IdentityKey,
        room_id: RoomID,
        forwarding_chain: Optional[List[IdentityKey]] = None,
    ) -> "InboundGroupSession":
        session = super().from_pickle(pickle, passphrase)
        session.signing_key = signing_key
        session.sender_key = sender_key
        session.room_id = room_id
        session.forwarding_chain = forwarding_chain or []
        return session

    @classmethod
    def import_session(
        cls,
        session_key: str,
        signing_key: SigningKey,
        sender_key: IdentityKey,
        room_id: RoomID,
        forwarding_chain: Optional[List[str]] = None,
    ) -> "InboundGroupSession":
        session = super().import_session(session_key)
        session.signing_key = signing_key
        session.sender_key = sender_key
        session.room_id = room_id
        session.forwarding_chain = forwarding_chain or []
        return session


class OutboundGroupSession(olm.OutboundGroupSession):
    """Outbound group session aware of the users it is shared with.

    Also remembers the time it was created and the number of messages it has
    encrypted, in order to know if it needs to be rotated.
    """

    max_age: timedelta
    max_messages: int
    creation_time: datetime
    use_time: datetime
    message_count: int

    room_id: RoomID
    users_shared_with: Set[Tuple[UserID, DeviceID]]
    users_ignored: Set[Tuple[UserID, DeviceID]]
    shared: bool

    def __init__(self, room_id: RoomID) -> None:
        self.max_age = timedelta(days=7)
        self.max_messages = 100
        self.creation_time = datetime.now()
        self.use_time = datetime.now()
        self.message_count = 0
        self.room_id = room_id
        self.users_shared_with = set()
        self.users_ignored = set()
        self.shared = False
        super().__init__()

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    @property
    def expired(self):
        return (
            self.message_count >= self.max_messages
            or datetime.now() - self.creation_time >= self.max_age
        )

    def encrypt(self, plaintext):
        if not self.shared:
            raise EncryptionError("Group session has not been shared")

        if self.expired:
            raise EncryptionError("Group session has expired")

        self.message_count += 1
        self.use_time = datetime.now()
        return super().encrypt(plaintext)

    @classmethod
    def from_pickle(
        cls,
        pickle: bytes,
        passphrase: str,
        max_age: timedelta,
        max_messages: int,
        creation_time: datetime,
        use_time: datetime,
        message_count: int,
        room_id: RoomID,
        shared: bool,
    ) -> "OutboundGroupSession":
        session = cast(OutboundGroupSession, super().from_pickle(pickle, passphrase))
        session.max_age = max_age
        session.max_messages = max_messages
        session.creation_time = creation_time
        session.use_time = use_time
        session.message_count = message_count
        session.room_id = room_id
        session.users_shared_with = set()
        session.users_ignored = set()
        session.shared = shared
        return session

    @property
    def share_content(self) -> RoomKeyEventContent:
        return RoomKeyEventContent(
            algorithm=EncryptionAlgorithm.MEGOLM_V1,
            room_id=self.room_id,
            session_id=self.id,
            session_key=self.session_key,
        )



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Callable
import asyncio
import os

from attr import dataclass

from mautrix.errors import MatrixConnectionError
from mautrix.types import UserID
from mautrix.util.manhole import start_manhole

from . import SECTION_ADMIN, CommandEvent, command_handler


@dataclass
class ManholeState:
    server: asyncio.AbstractServer
    opened_by: UserID
    close: Callable[[], None]
    whitelist: set[int]


@command_handler(
    needs_auth=False,
    needs_admin=True,
    help_section=SECTION_ADMIN,
    help_text="Open a manhole into the bridge.",
    help_args="<_uid..._>",
)
async def open_manhole(evt: CommandEvent) -> None:
    if not evt.config["manhole.enabled"]:
        await evt.reply("The manhole has been disabled in the config.")
        return
    elif len(evt.args) == 0:
        await evt.reply("**Usage:** `$cmdprefix+sp open-manhole <uid...>`")
        return

    whitelist = set()
    whitelist_whitelist = evt.config["manhole.whitelist"]
    for arg in evt.args:
        try:
            uid = int(arg)
        except ValueError:
            await evt.reply(f"{arg} is not an integer.")
            return
        if whitelist_whitelist and uid not in whitelist_whitelist:
            await evt.reply(f"{uid} is not in the list of allowed UIDs.")
            return
        whitelist.add(uid)

    if evt.bridge.manhole:
        added = [uid for uid in whitelist if uid not in evt.bridge.manhole.whitelist]
        evt.bridge.manhole.whitelist |= set(added)
        if len(added) == 0:
            await evt.reply(
                f"There's an existing manhole opened by {evt.bridge.manhole.opened_by}"
                " and all the given UIDs are already whitelisted."
            )
        else:
            added_str = (
                f"{', '.join(str(uid) for uid in added[:-1])} and {added[-1]}"
                if len(added) > 1
                else added[0]
            )
            await evt.reply(
                f"There's an existing manhole opened by {evt.bridge.manhole.opened_by}"
                f". Added {added_str} to the whitelist."
            )
            evt.log.info(f"{evt.sender.mxid} added {added_str} to the manhole whitelist.")
        return

    namespace = await evt.bridge.manhole_global_namespace(evt.sender.mxid)
    banner = evt.bridge.manhole_banner(evt.sender.mxid)
    path = evt.config["manhole.path"]

    wl_list = list(whitelist)
    whitelist_str = (
        f"{', '.join(str(uid) for uid in wl_list[:-1])} and {wl_list[-1]}"
        if len(wl_list) > 1
        else wl_list[0]
    )
    evt.log.info(f"{evt.sender.mxid} opened a manhole with {whitelist_str} whitelisted.")
    server, close = await start_manhole(
        path=path, banner=banner, namespace=namespace, loop=evt.loop, whitelist=whitelist
    )
    evt.bridge.manhole = ManholeState(
        server=server, opened_by=evt.sender.mxid, close=close, whitelist=whitelist
    )
    plrl = "s" if len(whitelist) != 1 else ""
    await evt.reply(f"Opened manhole at unix://{path} with UID{plrl} {whitelist_str} whitelisted")
    await server.wait_closed()
    evt.bridge.manhole = None
    try:
        os.unlink(path)
    except FileNotFoundError:
        pass
    evt.log.info(f"{evt.sender.mxid}'s manhole was closed.")
    try:
        await evt.reply("Your manhole was closed.")
    except (AttributeError, MatrixConnectionError) as e:
        evt.log.warning(f"Failed to send manhole close notification: {e}")


@command_handler(
    needs_auth=False,
    needs_admin=True,
    help_section=SECTION_ADMIN,
    help_text="Close an open manhole.",
)
async def close_manhole(evt: CommandEvent) -> None:
    if not evt.bridge.manhole:
        await evt.reply("There is no open manhole.")
        return

    opened_by = evt.bridge.manhole.opened_by
    evt.bridge.manhole.close()
    evt.bridge.manhole = None
    if opened_by != evt.sender.mxid:
        await evt.reply(f"Closed manhole opened by {opened_by}")



from .handler import (
    SECTION_ADMIN,
    SECTION_AUTH,
    SECTION_GENERAL,
    SECTION_RELAY,
    CommandEvent,
    CommandHandler,
    CommandHandlerFunc,
    CommandProcessor,
    HelpCacheKey,
    HelpSection,
    command_handler,
)
from .meta import cancel, help_cmd, unknown_command

from . import (  # isort: skip
    admin,
    clean_rooms,
    crypto,
    delete_portal,
    login_matrix,
    manhole,
    relay,
)

__all__ = [
    "HelpSection",
    "HelpCacheKey",
    "command_handler",
    "CommandHandler",
    "CommandProcessor",
    "CommandHandlerFunc",
    "CommandEvent",
    "SECTION_GENERAL",
    "SECTION_ADMIN",
    "SECTION_AUTH",
    "SECTION_RELAY",
]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from enum import Enum


class Scheme(Enum):
    POSTGRES = "postgres"
    COCKROACH = "cockroach"
    SQLITE = "sqlite"

    def __eq__(self, other: Scheme | str) -> bool:
        if isinstance(other, str):
            return self.value == other
        else:
            return super().__eq__(other)
